import inspect

import numpy as np


def get_array_ptr(data):
    try:
        return data.__cuda_array_interface__['data'][0]
    except RuntimeError:
        # Handle torch with gradient enabled
        # https://github.com/flatironinstitute/finufft/pull/326#issuecomment-1652212770
        return data.data_ptr()
    except AttributeError:
        raise TypeError("Invalid GPU array implementation. Implementation must implement the standard cuda array interface.")


def get_array_module(obj):
    module_name = inspect.getmodule(type(obj)).__name__

    if module_name.startswith("numba.cuda"):
        return "numba"
    elif module_name.startswith("torch"):
        return "torch"
    elif module_name.startswith("pycuda"):
        return "pycuda"
    else:
        return "generic"


def get_array_size(obj):
    array_module = get_array_module(obj)

    if array_module == "torch":
        return len(obj)
    else:
        return obj.size


def get_array_dtype(obj):
    array_module = get_array_module(obj)

    if array_module == "torch":
        dtype_str = str(obj.dtype)
        dtype_str = dtype_str[len("torch."):]
        return np.dtype(dtype_str)
    else:
        return obj.dtype


def is_array_contiguous(obj):
    array_module = get_array_module(obj)

    if array_module == "numba":
        return obj.is_c_contiguous()
    elif array_module == "torch":
        return obj.is_contiguous()
    else:
        return obj.flags.c_contiguous


def array_can_contiguous(obj):
    array_module = get_array_module(obj)

    if array_module == "pycuda":
        return False
    else:
        return True


def array_contiguous(obj):
    array_module = get_array_module(obj)

    if array_module == "numba":
        import numba
        ret = numba.cuda.device_array(obj.shape, obj.dtype, stream=obj.stream)
        ret[:] = obj[:]
        return ret
    if array_module == "torch":
        return obj.contiguous()
    else:
        return obj.copy(order="C")


def array_empty_like(obj, *args, **kwargs):
    module_name = get_array_module(obj)

    if module_name == "numba":
        import numba.cuda
        return numba.cuda.device_array(*args, **kwargs)
    elif module_name == "torch":
        import torch
        if "shape" in kwargs:
            kwargs["size"] = kwargs.pop("shape")
        if "dtype" in kwargs:
            dtype = kwargs.pop("dtype")
            if dtype == np.complex64:
                dtype = torch.complex64
            elif dtype == np.complex128:
                dtype = torch.complex128
            kwargs["dtype"] = dtype
        if "device" not in kwargs:
            kwargs["device"] = obj.device

        return torch.empty(*args, **kwargs)
    else:
        return type(obj)(*args, **kwargs)



# -*- coding: utf-8 -*-
#
# finufft documentation build configuration file, created by
# sphinx-quickstart on Wed Nov  1 16:19:13 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os
import sphinx.ext.autodoc

# autodoc py module stuff...

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('../python/finufft'))
sys.path.insert(0, os.path.abspath('../python/cufinufft'))

# quash sphinx's inevitable failure of C++ lib extension to import...
autodoc_mock_imports = ['finufft._finufft', 'numpy', 'pycuda', 'cufinufft._cufinufft']
# The above is not enough for nested import -- forcibly mock them out ahead of time:
#for name in autodoc_mock_imports:
#    sys.modules[name] = sphinx.ext.autodoc._MockModule(name, None)
# (removed the above since broke in sphinx 1.6 -> 1.7.4). Thanks Dylan.


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#needs_sphinx = '1.6'   # dylan
needs_sphinx = '1.3'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.todo',
    'sphinx.ext.napoleon',
    'sphinx.ext.mathjax',
#    'sphinx.ext.autosectionlabel',   # needs v 1.4; can :ref: other files w/o this; removed 7/29/18
    'texext',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'finufft'
copyright = u'2017-2024 The Simons Foundation, Inc. - All Rights Reserved'
author = u'Alex Barnett, Ludvig af Klinteberg, Joakim Andén, Libin Lu, Jeremy Magland, Andrea Malleo, Yu-Hsuan Shih, Marco Barbone, Robert Blackwell, and others.'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = u'2.3'
# The full version, including alpha/beta/rc tags.
release = u'2.3.1'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# Prevent readthedocs from highlighting :: blocks as py:
# https://stackoverflow.com/questions/44376893/selectively-disable-readthedocs-syntax-highlighting
highlight_language = 'none'

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files (.rst only)
exclude_patterns = ['_build', 'overview_meat']

# The reST default role (used for this markup: `text`) to use for all
# documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
#keep_warnings = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#html_theme = 'alabaster'
html_theme = 'classic'
#html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#html_theme_options = {}
#html_theme_options = {"codebgcolor":"rgb(240,240,240)"}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = 'logo-small.png'

# The name of an image file (relative to this directory) to use as a favicon of
# the docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# web browser icon. see SO: https://stackoverflow.com/questions/54639192/change-the-favicon-of-the-sphinx-read-the-docs-theme
html_favicon = 'logo-32x32.png'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
#html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
html_sidebars = {'**': ['globaltoc.html', 'relations.html', 'sourcelink.html', 'searchbox.html'] }

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
html_domain_indices = False

# If false, no index is generated.
html_use_index = False

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   'da', 'de', 'en', 'es', 'fi', 'fr', 'hu', 'it', 'ja'
#   'nl', 'no', 'pt', 'ro', 'ru', 'sv', 'tr'
#html_search_language = 'en'

# A dictionary with options for the search language support, empty by default.
# Now only 'ja' uses this config value
#html_search_options = {'type': 'default'}

# The name of a javascript file (relative to the configuration directory) that
# implements a search results scorer. If empty, the default will be used.
#html_search_scorer = 'scorer.js'

# Output file base name for HTML help builder.
htmlhelp_basename = 'finufftdoc'

# To fix location of equation numbering. Barnett tried 6/19/18
# see https://samnicholls.net/2016/06/15/how-to-sphinx-readthedocs/
def setup(app):
    app.add_css_file('theme_overrides.css')
# it doesn't fail if this file not found in _static  :(




# -- Options for LaTeX output ---------------------------------------------

# My problem with chapters wrongly numbered:
# https://stackoverflow.com/questions/60877841/how-can-i-force-sphinx-to-end-a-chapter-when-making-a-pdf
# .. ?

latex_elements = {
# The paper size ('letterpaper' or 'a4paper').
'papersize': 'letterpaper',

# The font size ('10pt', '11pt' or '12pt').
'pointsize': '10pt',

# removes blank pages:
'classoptions': ',openany,oneside',

# Additional stuff for the LaTeX preamble.
#'preamble': '',
'preamble': r'''
  \usepackage{hyperref}
  \setcounter{tocdepth}{3}
''',

# Latex figure (float) alignment
#'figure_align': 'htbp',

# prevent any index from being made (Komiya):
# https://groups.google.com/g/sphinx-users/c/Pn3uHevtR28
'makeindex': '',
'printindex': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    ('latexindex', 'finufft.tex', u'FINUFFT Documentation',
     'Alex Barnett and the FINUFFT team', 'manual',False),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
latex_logo = 'logo.png'

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
latex_domain_indices = False

# this doens't stop a latex index... :(
latex_use_modindex = False




# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'finufft', u'FINUFFT Documentation',
     [author], 1)
]

# If true, show URL addresses after external links.
#man_show_urls = False


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'finufft', u'FINUFFT Documentation',
     author, 'FINUFFT nonuniform FFT', 'Nonuniform FFT of types 1,2,3 in dimensions 1,2,3',
     'Numerical methods'),
]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
#texinfo_no_detailmenu = False



import numpy as np

import pycuda.autoinit
from pycuda.gpuarray import to_gpu

import cufinufft

# number of nonuniform points
M = 100000

# grid size
N = 200000

# generate positions for the nonuniform points and the coefficients
x = 2 * np.pi * np.random.uniform(size=M)
c = (np.random.standard_normal(size=M)
     + 1J * np.random.standard_normal(size=M))

# move the data to GPU
x_gpu = to_gpu(x)
c_gpu = to_gpu(c)

# compute the transform
f_gpu = cufinufft.nufft1d1(x_gpu, c_gpu, (N,))

# move results off the GPU
f = f_gpu.get()



import numpy as np


def _complex_dtype(dtype):
    dtype = np.dtype(dtype)

    if dtype == np.float32:
        complex_dtype = np.complex64
    elif dtype == np.float64:
        complex_dtype = np.complex128
    else:
        raise TypeError("dtype should be np.float32 or np.float64.")

    return complex_dtype


def _real_dtype(complex_dtype):
    complex_dtype = np.dtype(complex_dtype)

    if complex_dtype == np.complex64:
        real_dtype = np.float32
    elif complex_dtype == np.complex128:
        real_dtype = np.float64
    else:
        raise TypeError("dtype should be np.complex64 or np.complex128.")

    return real_dtype


def gen_nu_pts(M, dim=3, seed=0):
    np.random.seed(seed)
    k = np.random.uniform(-np.pi, np.pi, (dim, M))
    k = k.astype(np.float64)
    return k


def gen_uniform_data(shape, seed=0):
    np.random.seed(seed)
    fk = np.random.standard_normal(shape + (2,))
    fk = fk.astype(np.float64).view(np.complex128)[..., 0]
    return fk


def gen_nonuniform_data(M, seed=0, n_trans=()):
    np.random.seed(seed)
    c = np.random.standard_normal(2 * M * int(np.prod(n_trans)))
    c = c.astype(np.float64).view(np.complex128)
    c = c.reshape(n_trans + (M,))
    return c


def gen_sig_idx(n_modes, n_tr):
    idx = tuple(np.random.randint(0, n) for n in n_tr + n_modes)
    return idx


def gen_coef_ind(n_pts, n_tr):
    ind = tuple(np.random.randint(0, n) for n in n_tr + (n_pts,))
    return ind


def type1_problem(dtype, shape, M, n_trans=()):
    real_dtype = _real_dtype(dtype)
    dim = len(shape)

    k = gen_nu_pts(M, dim=dim).astype(real_dtype)
    c = gen_nonuniform_data(M, n_trans=n_trans).astype(dtype)

    return k, c


def type2_problem(dtype, shape, M, n_trans=()):
    real_dtype = _real_dtype(dtype)
    dim = len(shape)

    k = gen_nu_pts(M, dim=dim).astype(real_dtype)
    fk = gen_uniform_data(n_trans + shape).astype(dtype)

    return k, fk

def type3_problem(dtype, dim, n_source_pts, n_target_pts, n_trans=()):
    real_dtype = _real_dtype(dtype)

    source_pts = gen_nu_pts(n_source_pts, dim=dim).astype(real_dtype)
    source_coefs = gen_nonuniform_data(n_source_pts, n_trans=n_trans).astype(dtype)
    target_pts = gen_nu_pts(n_target_pts, dim=dim).astype(real_dtype)

    return source_pts, source_coefs, target_pts


def make_grid(shape):
    dim = len(shape)
    shape = shape

    grids = [np.arange(-(N // 2), (N + 1) // 2) for N in shape]
    grids = np.meshgrid(*grids, indexing='ij')
    return np.stack(grids)


def direct_type1(pts, coefs, n_modes, idx):
    dtype = coefs.dtype
    dim = len(n_modes)

    _idx = (np.array(idx[-dim:]) - np.floor(np.array(n_modes) / 2)).astype(dtype)

    _coefs = coefs[idx[:-dim]]

    sig = np.sum(np.exp(1j * np.sum(_idx[:, np.newaxis] * pts, axis=0)) * _coefs)

    return sig


def direct_type2(pts, sig, ind):
    dtype = sig.dtype
    dim = pts.shape[0]
    n_modes = sig.shape[-dim:]

    grids = [slice(-np.floor(n / 2), np.ceil(n / 2)) for n in n_modes]
    grid = np.mgrid[grids].astype(dtype)

    pt = pts[:, ind[-1]]
    pt = pt.reshape((dim,) + dim * (1,))

    _sig = sig[ind[:-1]]

    coef = np.sum(np.exp(-1j * np.sum(pt * grid, axis=0)) * _sig)

    return coef


def direct_type3(source_pts, source_coefs, target_pts, ind):
    target_pt = target_pts[:, ind[-1]]
    target_pt = target_pt[:, np.newaxis]

    _source_coef = source_coefs[ind[:-1]]

    target_coef = np.sum(np.exp(1j * np.sum(target_pt * source_pts, axis=0)) * _source_coef)

    return target_coef


def verify_type1(pts, coefs, shape, sig_est, tol):
    dim = pts.shape[0]

    n_trans = coefs.shape[:-1]

    assert sig_est.shape[:-dim] == n_trans
    assert sig_est.shape[-dim:] == shape

    idx = gen_sig_idx(shape, n_trans)

    fk_est = sig_est[idx]
    fk_target = direct_type1(pts, coefs, shape, idx)

    type1_rel_err = np.linalg.norm(fk_target - fk_est) / np.linalg.norm(fk_target)

    assert type1_rel_err < 25 * tol


def verify_type2(pts, sig, coefs_est, tol):
    dim = pts.shape[0]

    n_trans = sig.shape[:-dim]
    n_pts = pts.shape[-1]

    assert coefs_est.shape == n_trans + (n_pts,)

    ind = gen_coef_ind(n_pts, n_trans)

    c_est = coefs_est[ind]
    c_target = direct_type2(pts, sig, ind)

    type2_rel_err = np.linalg.norm(c_target - c_est) / np.linalg.norm(c_target)

    assert type2_rel_err < 25 * tol


def verify_type3(source_pts, source_coef, target_pts, target_coef, tol):
    dim = source_pts.shape[0]

    n_source_pts = source_pts.shape[-1]
    n_target_pts = target_pts.shape[-1]
    n_tr = source_coef.shape[:-1]

    assert target_coef.shape == n_tr + (n_target_pts,)

    ind = gen_coef_ind(n_source_pts, n_tr)

    target_est = target_coef[ind]
    target_true = direct_type3(source_pts, source_coef, target_pts, ind)

    type3_rel_err = np.linalg.norm(target_est - target_true) / np.linalg.norm(target_true)

    assert type3_rel_err < 100 * tol



import re


decimalMatchString = "\d+\.?\d+" #regular expression to match a decimal number
sciNotString = "(\d*.?\d*e-\d* s)" #regular expression to match a number in scientific notation
wholeNumberMatchString = "\d+" 


#search string needs to have two groupings! (one for everything besides) (time s)
def extractTime(searchString, strOut):
    time = 0
    lineMatch = re.search(searchString,strOut)
    if(lineMatch):
        val = re.search(sciNotString,lineMatch.group(2))
        if(not val):
            val = re.search(decimalMatchString, lineMatch.group(2))
        if(not val):
            val = re.search(wholeNumberMatchString, lineMatch.group(2))
        time = round(float(val.group(0).split('s')[0].strip()),5)
    return time


def sumAllTime(searchString, strOut):
    newVal = 0
    lineMatch = re.findall(searchString,strOut)
    for match in lineMatch:
        val = re.search(sciNotString, match[1])
        if(not val): #search failed, try decimal format 
            val = re.search(decimalMatchString, match[1])
        if(not val):
            val = re.search(wholeNumberMatchString, match[1])
        newVal = newVal + float(val.group(0).split('s')[0].strip()) #trim off " s"
    newVal = round(newVal,5)
    return newVal





import json

matrix = {
    "include": []
}

static_linking = ["On", "Off"]

combinations = [
    ("ubuntu-22.04", {
        "build_type": ["Release", "Debug"],
        "toolchain": ["llvm", "gcc"],
        "arch_flags": ["-march=native", "-march=x86-64", "native"],
        "ducc_fft": ["On", "Off"]
    }),
    ("windows-2022", {
        "build_type": ["Release", "Debug"],
        "toolchain": ["msvc"],
        "arch_flags": ["/arch:AVX2", "/arch:SSE2", "native"],
        "ducc_fft": ["On", "Off"]
    }),
    ("windows-2022", {
        "build_type": ["Release"],
        "toolchain": ["llvm"],
        "arch_flags": ["-march=native", "-march=x86-64", "native"],
        "ducc_fft": ["On", "Off"]
    }),
    ("macos-13", {
        "build_type": ["Release", "Debug"],
        "toolchain": ["llvm", "gcc-14"],
        "arch_flags": ["-march=native", "-march=x86-64", "native"],
        "ducc_fft": ["On", "Off"]
    })
]


def get_c_compiler(toolchain):
    if "gcc" in toolchain:
        return "gcc"
    elif toolchain == "llvm":
        return "clang"
    elif toolchain == "msvc":
        return "cl"
    else:
        raise ValueError(f"Unknown toolchain: {toolchain}")


def get_cxx_compiler(toolchain):
    if "gcc" in toolchain:
        return "g++"
    elif toolchain == "llvm":
        return "clang++"
    elif toolchain == "msvc":
        return "cl"
    else:
        raise ValueError(f"Unknown toolchain: {toolchain}")


for platform, value in combinations:
    for toolchain in value["toolchain"]:
        for arch_flag in value["arch_flags"]:
            for linking in static_linking:
                for build in value["build_type"]:
                    for ducc in value["ducc_fft"]:
                        matrix["include"].append({
                            "os": platform,
                            "toolchain": toolchain,
                            "arch_flags": arch_flag,
                            "finufft_static_linking": linking,
                            "build_type": build,
                            "c_compiler": get_c_compiler(toolchain),
                            "cxx_compiler": get_cxx_compiler(toolchain),
                            "ducc_fft": ducc
                        })
json_str = json.dumps(matrix, ensure_ascii=False)
print(json_str)



"""
Demonstrate the 3D type 2 NUFFT using cuFINUFFT with overlapping transfer/compute and
compare to other methods
"""

import numpy as np
import cupy
import cufinufft
import time

rng = np.random.Generator(np.random.PCG64(100))


pinned_memory_pool = cupy.cuda.PinnedMemoryPool()
cupy.cuda.set_pinned_memory_allocator(pinned_memory_pool.malloc)

def _pin_memory(array):
    mem = cupy.cuda.alloc_pinned_memory(array.nbytes)
    ret = np.frombuffer(mem, array.dtype, array.size).reshape(array.shape)
    ret[...] = array
    return ret

def test():
    # Set up parameters for problem.
    N = (128, 128, 160)             # Size of uniform grid
    M = 6136781                     # Number of nonuniform points
    n_transf = 1                    # Number of nuffts to perform per batch
    n_tot = 32                      # Number of batches to execute
    eps = 1e-5                      # Requested tolerance
    dtype = np.float32              # Datatype (real)
    complex_dtype = np.complex64    # Datatype (complex)

    # Generate coordinates of non-uniform points.
    x = 2 * np.pi * (rng.random(size=M, dtype=dtype) - 0.5)
    y = 2 * np.pi * (rng.random(size=M, dtype=dtype) - 0.5)
    z = 2 * np.pi * (rng.random(size=M, dtype=dtype) - 0.5)

    # Generate grid values.
    fk_all_naive = (rng.standard_normal((n_tot, n_transf, *N), dtype=dtype) + 1j * rng.standard_normal((n_tot, n_transf, *N), dtype=dtype))
    fk_all = _pin_memory(fk_all_naive)
    c_all_async = _pin_memory(np.zeros(shape=(n_tot, n_transf, M), dtype=complex_dtype))
    c_all_sync = _pin_memory(np.zeros(shape=(n_tot, n_transf, M), dtype=complex_dtype))
    c_all_naive = np.zeros(shape=(n_tot, n_transf, M), dtype=complex_dtype)

    # Initialize the plan and set the points.
    plan_stream = cupy.cuda.Stream(null=True)
    plan = cufinufft.Plan(2, N, n_transf, eps=eps, dtype=complex_dtype, gpu_kerevalmeth=1, gpu_stream=plan_stream.ptr)
    plan.setpts(cupy.array(x), cupy.array(y), cupy.array(z))

    # Using a simple front/back buffer approach. backbuffer is for DtoH transfers, and front for
    # execution and HtoD transfers
    front_stream, back_stream = (cupy.cuda.Stream(), cupy.cuda.Stream())
    front_fk_gpu = cupy.empty(fk_all[0].shape, fk_all[0].dtype)
    back_fk_gpu = cupy.empty(fk_all[0].shape, fk_all[0].dtype)
    front_c_gpu = cupy.empty(c_all_async[0].shape, c_all_async[0].dtype)
    back_c_gpu = cupy.empty(c_all_async[0].shape, c_all_async[0].dtype)
    front_plan = cufinufft.Plan(2, N, n_transf, eps=eps, dtype=complex_dtype, gpu_kerevalmeth=1, gpu_stream=front_stream.ptr)
    front_plan.setpts(cupy.array(x), cupy.array(y), cupy.array(z))
    back_plan = cufinufft.Plan(2, N, n_transf, eps=eps, dtype=complex_dtype, gpu_kerevalmeth=1, gpu_stream=back_stream.ptr)
    back_plan.setpts(cupy.array(x), cupy.array(y), cupy.array(z))

    # Run with async
    st = time.time()
    front_fk_gpu.set(fk_all[0], stream=front_stream)
    for i in range(n_tot):
        if i + 1 < n_tot:
            back_fk_gpu.set(fk_all[i + 1], stream=back_stream)

        front_plan.execute(front_fk_gpu, out=front_c_gpu).get(out=c_all_async[i], stream=front_stream)

        back_stream, front_stream = front_stream, back_stream
        back_plan, front_plan = front_plan, back_plan
        back_fk_gpu, front_fk_gpu = front_fk_gpu, back_fk_gpu
        back_c_gpu, front_c_gpu = front_c_gpu, back_c_gpu

    back_stream.synchronize()
    async_time = time.time() - st

    # Run with best-practice synchronous
    st = time.time()
    for i in range(n_tot):
        front_fk_gpu.set(fk_all[i])
        plan.execute(front_fk_gpu, out=front_c_gpu).get(out=c_all_sync[i])

    sync_time = time.time() - st

    # Run in a relatively naive way
    st = time.time()
    for i in range(n_tot):
        c_all_naive[i, :] = plan.execute(cupy.array(fk_all_naive[i])).get()[:]

    naive_time = time.time() - st

    assert(np.linalg.norm(c_all_sync - c_all_naive) == 0.0)
    assert(np.linalg.norm(c_all_async - c_all_naive) == 0.0)

    print(f"async timing: {async_time}")
    print(f"sync timing: {sync_time}")
    print(f"naive timing: {naive_time}")
    print(f"speedup (sync / async): {round(sync_time / async_time, 2)}")
    print(f"speedup (naive / sync): {round(naive_time / sync_time, 2)}")
    print(f"speedup (naive / async): {round(naive_time / async_time, 2)}")

    # Since plans carry raw stream pointers which aren't reference counted, we need to make
    # sure they're deleted before the stream objects that hold them. Otherwise, the stream
    # might be deleted before cufinufft can use it in the deletion routines. Manually clear
    # them out here.
    del plan, front_plan, back_plan

test()



#!/usr/bin/env python3

import math
import numpy as np
import re
import subprocess

import searchForTimeMetrics as stm

type = 1
n_trials = 1

#Small problem, sequential Multithreading
cmdStringOne_string = "./finufftGuru_test " + str(n_trials) + " "+ str(type) +" 3 1e1 1e1 1e1 1e4 1e-6 1 0" 
cmdStringOne_desc = "Small Problem, sequential Multithreaded"

#small problem, simulataneous single threaded, nested at end
cmdStringTwo_string = "./finufftGuru_test " + str(n_trials) + " "+ str(type) +" 3 1e1 1e1 1e1 1e4 1e-6 1 1" 
cmdStringTwo_desc = "Small Problem, single threaded sort in parallel"

#Large problem, sequential Multithreading
cmdStringThree_string = "./finufftGuru_test " + str(n_trials) + " "+ str(type) +" 3 1e2 1e2 1e2 1e7 1e-6 1 0" 
cmdStringThree_desc = "Large Problem, sequential Multithreaded"

#Large problem, simulataneous single threaded, nested at end
cmdStringFour_string = "./finufftGuru_test " + str(n_trials) + " "+ str(type) +" 3 1e2 1e2 1e2 1e7 1e-6 1 1" 
cmdStringFour_desc = "Large Problem, single threaded sort in parallel"

cmdStrings = [cmdStringOne_string, cmdStringTwo_string, cmdStringThree_string, cmdStringFour_string]
cmdDescs = [cmdStringOne_desc, cmdStringTwo_desc, cmdStringThree_desc, cmdStringFour_desc]

for i in range(len(cmdStrings)):
    cmdString = cmdStrings[i].split(' ')
    cmddesc = cmdDescs[i]
    print(cmddesc)
    print(cmdStrings[i])
    out =  subprocess.run([cmdString[0], cmdString[1], cmdString[2], cmdString[3],cmdString[4], cmdString[5],
                           cmdString[6], cmdString[7], cmdString[8], cmdString[8], cmdString[10]], 
                      stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    
    strOut = out.stdout.decode() #convert bytes to string
    #print(strOut) #uncomment to see debug output
    ###############################################################################
    #Gather Spreading Stats
    
    #spread (old) / [sort+spread]  (new)            
    newSort = stm.sumAllTime('(.*finufft_setpts.*sort)(.*)', strOut) #if type three, adds together both sorts

    #collect spreading if any
    newSpread = stm.extractTime('(.*finufft.*exec.*spread)(.*)' , strOut) 

    #collect interp if any
    newInterp = stm.extractTime('(.*finufft.*exec.*interp)(.*)',strOut) 

    #collect the spread timings for each trial of old
    totalOldSpread = stm.sumAllTime('(.*spread.*ier)(.*)', strOut) #gets spread AND unspread (i.e. interpolation) for all trials

    spreadRatio = round(totalOldSpread/(newSort + newSpread + newInterp),5)

    print("Total New Sort/Spread/Interp: " )
    print(round(newSort + newSpread + newInterp,5))
    print("Total Old Sort/Spread/Interp: ")
    print(totalOldSpread)
    print("Spreading Ratio: Old/New " )
    print(spreadRatio)
    print("***************************************************************")



import pytest

import numpy as np
from cufinufft import Plan

import utils


def test_multi_type1(framework, dtype=np.float32, shape=(16, 16, 16), M=4096, tol=1e-3):
    if framework == "pycuda":
        import pycuda.driver as drv
        import pycuda.gpuarray as gpuarray
    else:
        pytest.skip("Multi-GPU support only tested for pycuda")

    complex_dtype = utils._complex_dtype(dtype)

    drv.init()

    dev_count = drv.Device.count()

    if dev_count == 1:
        pytest.skip()

    devs = [drv.Device(dev_id) for dev_id in range(dev_count)]

    dim = len(shape)

    errs = []

    for dev_id, dev in enumerate(devs):
        ctx = dev.make_context()

        k = utils.gen_nu_pts(M, dim=dim).astype(dtype)
        c = utils.gen_nonuniform_data(M).astype(complex_dtype)

        k_gpu = gpuarray.to_gpu(k)
        c_gpu = gpuarray.to_gpu(c)
        fk_gpu = gpuarray.GPUArray(shape, dtype=complex_dtype)

        plan = Plan(1, shape, eps=tol, dtype=complex_dtype,
                         gpu_device_id=dev_id)

        plan.setpts(k_gpu[0], k_gpu[1], k_gpu[2])

        plan.execute(c_gpu, fk_gpu)

        fk = fk_gpu.get()

        ind = int(0.1789 * np.prod(shape))

        fk_est = fk.ravel()[ind]
        fk_target = utils.direct_type1(c, k, shape, ind)

        type1_rel_err = np.abs(fk_target - fk_est) / np.abs(fk_target)

        print(f'Type 1 relative error (GPU {dev_id}):', type1_rel_err)

        ctx.pop()

        errs.append(type1_rel_err)

    assert all(err < 0.01 for err in errs)



import pytest

import numpy as np

from cufinufft import Plan, _compat

import utils

# NOTE: Tests below fail for tolerance 1e-4 (error executing plan).

DTYPES = [np.float32, np.float64]
SHAPES = [(16,), (16, 16), (16, 16, 16), (19,), (17, 19), (17, 19, 24)]
MS = [256, 1024, 4096]
TOLS = [1e-3, 1e-6]
OUTPUT_ARGS = [False, True]
CONTIGUOUS = [False, True]
MODEORDS = [0, 1]


@pytest.mark.parametrize("dtype", DTYPES)
@pytest.mark.parametrize("shape", SHAPES)
@pytest.mark.parametrize("M", MS)
@pytest.mark.parametrize("tol", TOLS)
@pytest.mark.parametrize("output_arg", OUTPUT_ARGS)
@pytest.mark.parametrize("modeord", MODEORDS)
def test_type1(to_gpu, to_cpu, dtype, shape, M, tol, output_arg, modeord):
    complex_dtype = utils._complex_dtype(dtype)

    k, c = utils.type1_problem(dtype, shape, M)

    k_gpu = to_gpu(k)
    c_gpu = to_gpu(c)

    plan = Plan(1, shape, eps=tol, dtype=complex_dtype, modeord=modeord)

    # Since k_gpu is an array of shape (dim, M), this will expand to
    # plan.setpts(k_gpu[0], ..., k_gpu[dim]), allowing us to handle all
    # dimensions with the same call.
    plan.setpts(*k_gpu)

    if output_arg:
        fk_gpu = _compat.array_empty_like(c_gpu, shape, dtype=complex_dtype)
        plan.execute(c_gpu, out=fk_gpu)
    else:
        fk_gpu = plan.execute(c_gpu)

    fk = to_cpu(fk_gpu)
    if modeord == 1:
        fk = np.fft.fftshift(fk)

    utils.verify_type1(k, c, fk, tol)


@pytest.mark.parametrize("dtype", DTYPES)
@pytest.mark.parametrize("shape", SHAPES)
@pytest.mark.parametrize("M", MS)
@pytest.mark.parametrize("tol", TOLS)
@pytest.mark.parametrize("output_arg", OUTPUT_ARGS)
@pytest.mark.parametrize("contiguous", CONTIGUOUS)
@pytest.mark.parametrize("modeord", MODEORDS)
def test_type2(to_gpu, to_cpu, dtype, shape, M, tol, output_arg, contiguous, modeord):
    complex_dtype = utils._complex_dtype(dtype)

    k, fk = utils.type2_problem(dtype, shape, M)

    plan = Plan(2, shape, eps=tol, dtype=complex_dtype, modeord=modeord)

    check_result = True

    if not contiguous and len(shape) > 1:
        fk = fk.copy(order="F")

        if _compat.array_can_contiguous(to_gpu(np.empty(1))):
            def _execute(*args, **kwargs):
                with pytest.warns(UserWarning, match="requirement: C. Copying"):
                    return plan.execute(*args, **kwargs)
        else:
            check_result = False

            def _execute(*args, **kwargs):
                with pytest.raises(TypeError, match="requirement: C"):
                    plan.execute(*args, **kwargs)

    else:
        def _execute(*args, **kwargs):
            return plan.execute(*args, **kwargs)

    k_gpu = to_gpu(k)

    if modeord == 1:
        _fk = np.fft.ifftshift(fk)
    else:
        _fk = fk
    fk_gpu = to_gpu(_fk)

    plan.setpts(*k_gpu)

    if output_arg:
        c_gpu = _compat.array_empty_like(fk_gpu, (M,), dtype=complex_dtype)
        _execute(fk_gpu, out=c_gpu)
    else:
        c_gpu = _execute(fk_gpu)

    if check_result:
        c = to_cpu(c_gpu)

        utils.verify_type2(k, fk, c, tol)


def test_opts(to_gpu, to_cpu, shape=(8, 8, 8), M=32, tol=1e-3):
    dtype = np.float32

    complex_dtype = utils._complex_dtype(dtype)

    k, c = utils.type1_problem(dtype, shape, M)

    k_gpu = to_gpu(k)
    c_gpu = to_gpu(c)
    fk_gpu = _compat.array_empty_like(c_gpu, shape, dtype=complex_dtype)

    plan = Plan(1, shape, eps=tol, dtype=complex_dtype, gpu_sort=False,
                     gpu_maxsubprobsize=10)

    plan.setpts(k_gpu[0], k_gpu[1], k_gpu[2])

    plan.execute(c_gpu, fk_gpu)

    fk = to_cpu(fk_gpu)

    utils.verify_type1(k, c, fk, tol)



from cufinufft._plan import Plan

from cufinufft._simple import (nufft1d1, nufft1d2, nufft2d1, nufft2d2,
                               nufft3d1, nufft3d2)

__all__ = ["nufft1d1", "nufft1d2",
           "nufft2d1", "nufft2d2",
           "nufft3d1", "nufft3d2",
           "Plan"]

__version__ = '2.3.1'



# Simple 1d1 python interface call
# Lu 02/07/20.

import time
import finufft as fp
import numpy as np

np.random.seed(42)

N = int(1e6)
M = int(1e5)
x = np.random.uniform(-np.pi, np.pi, M)
c = np.random.randn(M) + 1.j * np.random.randn(M)
F = np.zeros([N], dtype=np.complex128)       # allocate F (modes out)
n_modes = np.ones([1], dtype=np.int64)
n_modes[0] = N

strt = time.time()

#plan
plan = fp.Plan(1,(N,))

#set pts
plan.setpts(x)

#exec
plan.execute(c,F)

#timing
print("Finished nufft in {0:.2g} seconds. Checking..."
      .format(time.time()-strt))

#check error
n = 142519      # mode to check
Ftest = 0.0
# this is so slow...
for j in range(M):
    Ftest += c[j] * np.exp(n * x[j] * 1.j)
Fmax = np.max(np.abs(F))
err = np.abs((F[n + N // 2] - Ftest) / Fmax)
print("Error relative to max of F: {0:.2e}".format(err))



import pytest

import numpy as np
from ctypes.util import find_library


# Check to make sure the fallback mechanism works if there is no bundled
# dynamic library.
@pytest.mark.skip(reason="Patching seems to fail in CI")
def test_fallback(mocker):
    def fake_load_library(lib_name, path):
        if lib_name in ["libcufinufft", "cufinufft"]:
            raise OSError()
        else:
            return np.ctypeslib.load_library(lib_name, path)

    # Block out the bundled library.
    mocker.patch("numpy.ctypeslib.load_library", fake_load_library)

    # Make sure an error is raised if no system library is found.
    if find_library("cufinufft") is None:
        with pytest.raises(ImportError, match="suitable cufinufft"):
            import cufinufft
    else:
        import cufinufft



#!/usr/bin/env python
"""
This module contains the high level python wrapper for
the cufinufft CUDA libraries.
"""

import atexit
import collections.abc
import numbers
import sys
import warnings

import numpy as np

from ctypes import byref
from ctypes import c_int64
from ctypes import c_void_p

from cufinufft._cufinufft import NufftOpts
from cufinufft._cufinufft import _default_opts
from cufinufft._cufinufft import _make_plan
from cufinufft._cufinufft import _make_planf
from cufinufft._cufinufft import _set_pts
from cufinufft._cufinufft import _set_ptsf
from cufinufft._cufinufft import _exec_plan
from cufinufft._cufinufft import _exec_planf
from cufinufft._cufinufft import _destroy_plan
from cufinufft._cufinufft import _destroy_planf

from cufinufft import _compat


# If we are shutting down python, we don't need to run __del__
#   This will avoid any shutdown gc ordering problems.
exiting = False
atexit.register(setattr, sys.modules[__name__], 'exiting', True)


class Plan:
    """
    A non-uniform fast Fourier transform (NUFFT) plan

    The ``Plan`` class lets the user exercise more fine-grained control over
    the execution of an NUFFT. First, the plan is created with a certain set
    of parameters (type, mode configuration, precision, sign, number of
    simultaneous transforms, and so on). Then the nonuniform points are set
    (source or target depending on the type). Finally, the plan is executed on
    some data, yielding the desired output.

    Args:
        nufft_type      (int): type of NUFFT (1 or 2).
        n_modes         (tuple of ints): the number of modes in each
                        dimension (for example `(50, 100)`).
        n_trans         (int, optional): number of transforms to compute.
        eps             (float, optional): precision requested (>1e-16).
        isign           (int, optional): if +1, uses the positive sign
                        exponential, otherwise the negative sign; defaults to
                        +1 for type 1 and to -1 for type 2.
        dtype           (string, optional): the precision of the transform,
                        'complex64' or 'complex128'.
        **kwargs        (optional): additional options corresponding may be
                        specified as keyword arguments. These include
                        upsampling ratio ``upsampfac`` (default ``2.0``),
                        ``gpu_method`` (1: nonuniform points-driven, 2: shared
                        memory), ``gpu_sort`` (for ``gpu_method == 1``, 0: no
                        sort, 1: sort), ``gpu_kerevalmeth`` (0: direct
                        exp(sqrt), Horner evaluation), ``gpu_device_id`` (GPU
                        ID), ``gpu_stream`` (CUDA stream pointer) and
                        ``modeord`` (0: CMCL-compatible mode ordering,
                        1: FFT-style mode ordering).
    """

    def __init__(self, nufft_type, n_modes, n_trans=1, eps=1e-6, isign=None,
                 dtype="complex64", **kwargs):
        if isign is None:
            if nufft_type == 2:
                isign = -1
            else:
                isign = +1

        # Need to set the plan here in case something goes wrong later on,
        # otherwise we error during __del__.
        self._plan = None

        # Setup type bound methods
        self.dtype = np.dtype(dtype)

        if self.dtype == np.float64:
            warnings.warn("Real dtypes are currently deprecated and will be "
                          "removed in version 2.3. Converting to complex128.",
                          DeprecationWarning)
            self.dtype = np.complex128

        if self.dtype == np.float32:
            warnings.warn("Real dtypes are currently deprecated and will be "
                          "removed in version 2.3. Converting to complex64.",
                          DeprecationWarning)
            self.dtype = np.complex64

        if self.dtype == np.complex128:
            self._make_plan = _make_plan
            self._setpts = _set_pts
            self._exec_plan = _exec_plan
            self._destroy_plan = _destroy_plan
            self.real_dtype = np.float64
        elif self.dtype == np.complex64:
            self._make_plan = _make_planf
            self._setpts = _set_ptsf
            self._exec_plan = _exec_planf
            self._destroy_plan = _destroy_planf
            self.real_dtype = np.float32
        else:
            raise TypeError("Expected complex64 or complex128.")

        if isinstance(n_modes, numbers.Integral):
            n_modes = (n_modes,)
        elif isinstance(n_modes, collections.abc.Iterable):
            n_modes = tuple(n_modes)
        else:
            raise ValueError(f"Invalid n_modes '{n_modes}'")

        self.dim = len(n_modes)
        self.type = nufft_type
        self.isign = isign
        self.eps = float(eps)
        self.n_modes = n_modes
        self.n_trans = n_trans
        self._maxbatch = 1    # TODO: optimize this one day

        # Get the default option values.
        self._opts = self._default_opts()

        # Extract list of valid field names.
        field_names = [name for name, _ in self._opts._fields_]

        # Assign field names from kwargs if they match up, otherwise error.
        for k, v in kwargs.items():
            if k in field_names:
                setattr(self._opts, k, v)
            else:
                raise TypeError(f"Invalid option '{k}'")

        # Initialize the plan.
        self._init_plan()

        # Initialize a list for references to objects
        #   we want to keep around for life of instance.
        self._references = []

    @staticmethod
    def _default_opts():
        """
        Generates a cufinufft opt struct of the dtype coresponding to plan.

        :param nufft_type: integer 1, 2, or 3.
        :param dim: Integer dimension.

        :return: nufft_opts structure.
        """

        nufft_opts = NufftOpts()

        _default_opts(nufft_opts)

        return nufft_opts

    def _init_plan(self):
        """
        Internal method to initialize plan struct and call low level make_plan.
        """

        # Initialize struct
        self._plan = c_void_p(None)

        # We extend the mode tuple to 3D as needed,
        #   and reorder from C/python ndarray.shape style input (nZ, nY, nX)
        #   to the (F) order expected by the low level library (nX, nY, nZ).
        _n_modes = self.n_modes[::-1] + (1,) * (3 - self.dim)
        _n_modes = (c_int64 * 3)(*_n_modes)

        ier = self._make_plan(self.type,
                              self.dim,
                              _n_modes,
                              self.isign,
                              self.n_trans,
                              self.eps,
                              byref(self._plan),
                              self._opts)

        if ier != 0:
            raise RuntimeError('Error creating plan.')

    def setpts(self, x, y=None, z=None, s=None, t=None, u=None):
        """
        Set the nonuniform points

        For type 1, this sets the coordinates of the ``M`` nonuniform source
        points and for type 2, it sets the coordinates of the ``M`` target
        points.

        The dimension of the plan determines the number of arguments supplied.
        For example, if ``dim == 2``, we provide ``x`` and ``y``.

        Args:
            x       (float[M]): first coordinate of the nonuniform points
                    (source for type 1, target for type 2).
            y       (float[M], optional): second coordinate of the nonuniform
                    points (source for type 1, target for type 2).
            z       (float[M], optional): third coordinate of the nonuniform
                    points (source for type 1, target for type 2).
        """

        _x = _ensure_array_type(x, "x", self.real_dtype)
        _y = _ensure_array_type(y, "y", self.real_dtype)
        _z = _ensure_array_type(z, "z", self.real_dtype)

        _x, _y, _z = _ensure_valid_pts(_x, _y, _z, self.dim)

        M = _compat.get_array_size(_x)

        # Because FINUFFT/cufinufft are internally column major,
        #   we will reorder the pts axes. Reordering references
        #   save us from having to actually transpose signal data
        #   from row major (Python default) to column major.
        #   We do this by following translation:
        #     (x, None, None) ~>  (x, None, None)
        #     (x, y, None)    ~>  (y, x, None)
        #     (x, y, z)       ~>  (z, y, x)
        # Via code, we push each dimension onto a stack of axis
        fpts_axes = [_compat.get_array_ptr(_x), None, None]

        # We will also store references to these arrays.
        #   This keeps python from prematurely cleaning them up.
        self._references.append(_x)
        if self.dim >= 2:
            fpts_axes.insert(0, _compat.get_array_ptr(_y))
            self._references.append(_y)

        if self.dim >= 3:
            fpts_axes.insert(0, _compat.get_array_ptr(_z))
            self._references.append(_z)

        # Then take three items off the stack as our reordered axis.
        ier = self._setpts(self._plan, M, *fpts_axes[:3], 0, None, None, None)

        self.nj = M

        if ier != 0:
            raise RuntimeError('Error setting non-uniform points.')

    def execute(self, data, out=None):
        """
        Execute the plan

        Performs the NUFFT specified at plan instantiation with the points set
        by ``setpts``. For type-1 and type-3 transforms, the input is a set of
        source strengths, while for a type-2 transform, it consists of an
        array of size ``n_modes``. If ``n_trans`` is greater than one,
        ``n_trans`` inputs are expected, stacked along the first axis.

        Args:
            data    (complex[M], complex[n_tr, M], complex[n_modes], or complex[n_tr, n_modes]):
                    The input source strengths (type 1) or source modes (type
                    2).
            out     (complex[n_modes], complex[n_tr, n_modes], complex[M], or complex[n_tr, M], optional):
                    The array where the output is stored. Must be of the
                    correct size.

        Returns:
            complex[n_modes], complex[n_tr, n_modes], complex[M], or complex[n_tr, M]:
            The output array of the transform(s).
        """

        _data = _ensure_array_type(data, "data", self.dtype)
        _out = _ensure_array_type(out, "out", self.dtype, output=True)

        if self.type == 1:
            req_data_shape = (self.n_trans, self.nj)
            req_out_shape = self.n_modes
        elif self.type == 2:
            req_data_shape = (self.n_trans, *self.n_modes)
            req_out_shape = (self.nj,)

        _data, data_shape = _ensure_array_shape(_data, "data", req_data_shape,
                                                allow_reshape=True)
        if self.type == 1:
            batch_shape = data_shape[:-1]
        else:
            batch_shape = data_shape[:-self.dim]

        req_out_shape = batch_shape + req_out_shape

        if out is None:
            _out = _compat.array_empty_like(_data, req_out_shape, dtype=self.dtype)
        else:
            _out = _ensure_array_shape(_out, "out", req_out_shape)

        if self.type == 1:
            ier = self._exec_plan(self._plan, _compat.get_array_ptr(_data),
                    _compat.get_array_ptr(_out))
        elif self.type == 2:
            ier = self._exec_plan(self._plan, _compat.get_array_ptr(_out),
                    _compat.get_array_ptr(_data))

        if ier != 0:
            raise RuntimeError('Error executing plan.')

        return _out

    def __del__(self):
        """
        Destroy this instance's associated plan and storage.
        """

        # If the process is exiting or we've already cleaned up plan, return.
        if exiting or self._plan is None:
            return

        ier = self._destroy_plan(self._plan)

        if ier != 0:
            raise RuntimeError('Error destroying plan.')

        # Reset plan to avoid double destroy.
        self._plan = None

        # Reset our reference.
        self._references = []


def _ensure_array_type(x, name, dtype, output=False):
    if x is None:
        return None

    if _compat.get_array_dtype(x) != dtype:
        raise TypeError(f"Argument `{name}` does not have the correct dtype: "
                        f"{x.dtype} was given, but {dtype} was expected.")

    if not _compat.is_array_contiguous(x):
        if output or not _compat.array_can_contiguous(x):
            raise TypeError(f"Argument `{name}` does not satisfy the "
                            f"following requirement: C")
        else:
            warnings.warn(f"Argument `{name}` does not satisfy the "
                          f"following requirement: C. Copying array "
                          f"(this may reduce performance)")
            x = _compat.array_contiguous(x)

    return x


def _ensure_array_shape(x, name, shape, allow_reshape=False):
    orig_shape = x.shape

    if x.shape != shape:
        if not allow_reshape or np.prod(x.shape) != np.prod(shape):
            raise TypeError(f"Argument `{name}` must be of shape {shape}")
        else:
            x = x.reshape(shape)

    if allow_reshape:
        return x, orig_shape
    else:
        return x


def _ensure_valid_pts(x, y, z, dim):
    if x.ndim != 1:
        raise TypeError(f"Argument `x` must be a vector")

    if dim >= 2:
        y = _ensure_array_shape(y, "y", x.shape)

    if dim >= 3:
        z = _ensure_array_shape(z, "z", x.shape)

    if dim < 3 and z is not None and _compat.get_array_size(z) > 0:
        raise TypeError(f"Plan dimension is {dim}, but `z` was specified")

    if dim < 2 and y is not None and _compat.get_array_size(y) > 0:
        raise TypeError(f"Plan dimension is {dim}, but `y` was specified")

    return x, y, z



import cupy as cp

import cufinufft

# number of nonuniform points
M = 100000

# grid size
N = 200000

# generate positions for the nonuniform points and the coefficients
x_gpu = 2 * cp.pi * cp.random.uniform(size=M)
c_gpu = (cp.random.standard_normal(size=M)
         + 1J * cp.random.standard_normal(size=M))

# compute the transform
f_gpu = cufinufft.nufft1d1(x_gpu, c_gpu, (N,))

# move results off the GPU
f = f_gpu.get()



import numpy as np


def _complex_dtype(dtype):
    if dtype == np.float32:
        complex_dtype = np.complex64
    elif dtype == np.float64:
        complex_dtype = np.complex128
    else:
        raise TypeError("dtype should be np.float32 or np.float64.")

    return complex_dtype


def _real_dtype(complex_dtype):
    if complex_dtype == np.complex64:
        real_dtype = np.float32
    elif complex_dtype == np.complex128:
        real_dtype = np.float64
    else:
        raise TypeError("dtype should be np.complex64 or np.complex128.")

    return real_dtype


def gen_nu_pts(M, dim=3, seed=0):
    np.random.seed(seed)
    k = np.random.uniform(-np.pi, np.pi, (dim, M))
    k = k.astype(np.float64)
    return k


def gen_uniform_data(shape, seed=0):
    np.random.seed(seed)
    fk = np.random.standard_normal(shape + (2,))
    fk = fk.astype(np.float64).view(np.complex128)[..., 0]
    return fk


def gen_nonuniform_data(M, seed=0, n_trans=()):
    np.random.seed(seed)
    c = np.random.standard_normal(2 * M * int(np.prod(n_trans)))
    c = c.astype(np.float64).view(np.complex128)
    c = c.reshape(n_trans + (M,))
    return c


def type1_problem(dtype, shape, M, n_trans=()):
    complex_dtype = _complex_dtype(dtype)
    dim = len(shape)

    k = gen_nu_pts(M, dim=dim).astype(dtype)
    c = gen_nonuniform_data(M, n_trans=n_trans).astype(complex_dtype)

    return k, c


def type2_problem(dtype, shape, M, n_trans=()):
    complex_dtype = _complex_dtype(dtype)
    dim = len(shape)

    k = gen_nu_pts(M, dim=dim).astype(dtype)
    fk = gen_uniform_data(n_trans + shape).astype(complex_dtype)

    return k, fk


def make_grid(shape):
    dim = len(shape)
    shape = shape

    grids = [np.arange(-(N // 2), (N + 1) // 2) for N in shape]
    grids = np.meshgrid(*grids, indexing='ij')
    return np.stack(grids)


def direct_type1(c, k, shape, ind):
    dim = len(shape)

    grid = make_grid(shape)

    phase = k.T @ grid.reshape((dim, -1))[:, ind]
    fk = np.sum(c * np.exp(1j * phase), -1)

    return fk


def direct_type2(fk, k, dim):
    grid = make_grid(fk.shape[-dim:])

    phase = k @ grid.reshape((dim, -1))

    # Ravel the spatial dimensions only.
    fk = fk.reshape(fk.shape[:-dim] + (np.prod(fk.shape[-dim:]),))
    c = np.sum(fk * np.exp(-1j * phase), -1)

    return c


def verify_type1(k, c, fk, tol):
    dim = fk.ndim - (c.ndim - 1)

    shape = fk.shape[-dim:]

    ind = int(0.1789 * np.prod(shape))

    # Ravel the spatial dimensions only.
    fk_est = fk.reshape(fk.shape[:-dim] + (np.prod(fk.shape[-dim:]),))[..., ind]
    fk_target = direct_type1(c, k, shape, ind)

    type1_rel_err = np.linalg.norm(fk_target - fk_est) / np.linalg.norm(fk_target)

    assert type1_rel_err < 25 * tol


def verify_type2(k, fk, c, tol):
    dim = fk.ndim - (c.ndim - 1)

    M = c.shape[-1]

    ind = M // 2

    c_est = c[..., ind]
    c_target = direct_type2(fk, k[:, ind], dim)

    type2_rel_err = np.linalg.norm(c_target - c_est) / np.linalg.norm(c_target)

    assert type2_rel_err < 25 * tol


def transfer_funcs(module_name):
    if module_name == "pycuda":
        import pycuda.autoinit # NOQA:401
        from pycuda.gpuarray import to_gpu
        def to_cpu(obj):
            return obj.get()
    elif module_name == "cupy":
        import cupy
        def to_gpu(obj):
            return cupy.array(obj)
        def to_cpu(obj):
            return obj.get()
    elif module_name == "numba":
        import numba.cuda
        to_gpu = numba.cuda.to_device
        def to_cpu(obj):
            return obj.copy_to_host()
    elif module_name == "torch":
        import torch
        def to_gpu(obj):
            return torch.as_tensor(obj, device=torch.device("cuda"))
        def to_cpu(obj):
            return obj.cpu().numpy()
    else:
        raise TypeError(f"Unsupported framework: {module_name}")

    return to_gpu, to_cpu



import pytest

import numpy as np

from finufft import Plan

import utils


# NOTE: Only doing single transformations for now.
SHAPES = [(7,), (8,), (7, 7), (7, 8), (8, 8), (7, 7, 7), (7, 8, 8), (8, 8, 8)]
N_PTS = [10, 11]
DTYPES = [np.complex64, np.complex128]
OUTPUT_ARGS = [False, True]
MODEORDS = [0, 1]


@pytest.mark.parametrize("dtype", DTYPES)
@pytest.mark.parametrize("shape", SHAPES)
@pytest.mark.parametrize("n_pts", N_PTS)
@pytest.mark.parametrize("output_arg", OUTPUT_ARGS)
@pytest.mark.parametrize("modeord", MODEORDS)
def test_finufft1_plan(dtype, shape, n_pts, output_arg, modeord):
    pts, coefs = utils.type1_problem(dtype, shape, n_pts)

    plan = Plan(1, shape, dtype=dtype, modeord=modeord)

    plan.setpts(*pts)

    if not output_arg:
        sig = plan.execute(coefs)
    else:
        sig = np.empty(shape, dtype=dtype)
        plan.execute(coefs, out=sig)

    if modeord == 1:
        sig = np.fft.fftshift(sig)

    utils.verify_type1(pts, coefs, shape, sig, 1e-6)


@pytest.mark.parametrize("dtype", DTYPES)
@pytest.mark.parametrize("shape", SHAPES)
@pytest.mark.parametrize("n_pts", N_PTS)
@pytest.mark.parametrize("output_arg", OUTPUT_ARGS)
@pytest.mark.parametrize("modeord", MODEORDS)
def test_finufft2_plan(dtype, shape, n_pts, output_arg, modeord):
    pts, sig = utils.type2_problem(dtype, shape, n_pts)

    plan = Plan(2, shape, dtype=dtype, modeord=modeord)

    plan.setpts(*pts)

    if modeord == 1:
        _sig = np.fft.ifftshift(sig)
    else:
        _sig = sig

    if not output_arg:
        coefs = plan.execute(_sig)
    else:
        coefs = np.empty(n_pts, dtype=dtype)
        plan.execute(_sig, out=coefs)

    utils.verify_type2(pts, sig, coefs, 1e-6)


@pytest.mark.parametrize("dtype", DTYPES)
@pytest.mark.parametrize("dim", list(set(len(shape) for shape in SHAPES)))
@pytest.mark.parametrize("n_source_pts", N_PTS)
@pytest.mark.parametrize("n_target_pts", N_PTS)
@pytest.mark.parametrize("output_arg", OUTPUT_ARGS)
def test_finufft3_plan(dtype, dim, n_source_pts, n_target_pts, output_arg):
    source_pts, source_coefs, target_pts = utils.type3_problem(dtype,
            dim, n_source_pts, n_target_pts)

    plan = Plan(3, dim, dtype=dtype)

    plan.setpts(*source_pts, *((None,) * (3 - dim)), *target_pts)

    if not output_arg:
        target_coefs = plan.execute(source_coefs)
    else:
        target_coefs = np.empty(n_target_pts, dtype=dtype)
        plan.execute(source_coefs, out=target_coefs)

    utils.verify_type3(source_pts, source_coefs, target_pts, target_coefs, 1e-6)


def test_finufft_plan_errors():
    with pytest.raises(RuntimeError, match="must be single or double"):
        Plan(1, (8, 8), dtype="uint32")

    with pytest.warns(Warning, match="finufft_opts does not have"):
        Plan(1, (8, 8), foo="bar")

    with pytest.raises(RuntimeError, match="type 3 plan n_modes_or_dim must be"):
        Plan(3, (1, 2))

    with pytest.raises(RuntimeError, match="n_modes dimension must be 1, 2"):
        Plan(2, (1, 2, 3, 4))

    with pytest.warns(Warning, match="eps tolerance too small"):
        Plan(1, (8, 8), eps=1e-30)

    with pytest.raises(TypeError, match="does not have the correct dtype"):
        Plan(1, (8, 8), dtype="complex64").setpts(np.ones(1, dtype="complex128"))

    with pytest.raises(TypeError, match="the following requirement: C"):
        plan = Plan(1, (8, 8), dtype="complex64")
        plan.setpts(*np.ones((2, 1), dtype="float32"))
        plan.execute(np.ones(1, dtype="complex64"), out=np.ones((8, 8),
            dtype="complex64", order="F"))

    with pytest.raises(TypeError, match="the following requirement: W"):
        plan = Plan(1, (8, 8), dtype="complex64")
        plan.setpts(*np.ones((2, 1), dtype="float32"))
        out = np.ones((8, 8), dtype="complex64")
        out.setflags(write=False)
        plan.execute(np.ones(1, dtype="complex64"), out=out)

    with pytest.warns(Warning, match="the following requirement: C. Copying"):
        plan = Plan(2, (8, 8), dtype="complex64")
        plan.setpts(*np.ones((2, 1), dtype="float32"))
        plan.execute(np.ones((8, 8), dtype="complex64", order="F"))

    vec = np.ones(1, dtype="float32")
    not_vec = np.ones((2, 1), dtype="float32")

    with pytest.raises(RuntimeError, match="x must be a vector"):
        Plan(1, (8, 8, 8), dtype="complex64").setpts(not_vec, vec, vec)

    with pytest.raises(RuntimeError, match="y must be a vector"):
        Plan(1, (8, 8, 8), dtype="complex64").setpts(vec, not_vec, vec)

    with pytest.raises(RuntimeError, match="z must be a vector"):
        Plan(1, (8, 8, 8), dtype="complex64").setpts(vec, vec, not_vec)

    with pytest.raises(RuntimeError, match="s must be a vector"):
        Plan(3, 3, dtype="complex64").setpts(vec, vec, vec, not_vec, vec, vec)

    with pytest.raises(RuntimeError, match="t must be a vector"):
        Plan(3, 3, dtype="complex64").setpts(vec, vec, vec, vec, not_vec, vec)

    with pytest.raises(RuntimeError, match="u must be a vector"):
        Plan(3, 3, dtype="complex64").setpts(vec, vec, vec, vec, vec, not_vec)

    vec = np.ones(3, dtype="float32")
    long_vec = np.ones(4, dtype="float32")

    with pytest.raises(RuntimeError, match="y must have same length as x"):
        Plan(1, (8, 8, 8), dtype="complex64").setpts(vec, long_vec, vec)

    with pytest.raises(RuntimeError, match="z must have same length as x"):
        Plan(1, (8, 8, 8), dtype="complex64").setpts(vec, vec, long_vec)

    with pytest.raises(RuntimeError, match="t must have same length as s"):
        Plan(3, 3, dtype="complex64").setpts(vec, vec, vec, vec, long_vec, vec)

    with pytest.raises(RuntimeError, match="u must have same length as s"):
        Plan(3, 3, dtype="complex64").setpts(vec, vec, vec, vec, vec, long_vec)

    with pytest.raises(RuntimeError, match="c.ndim must be 1 or 2 if n_trans == 1"):
        plan = Plan(1, (8,), dtype="complex64")
        plan.setpts(np.ones(3, dtype="float32"))
        plan.execute(np.ones((2, 1, 1), dtype="complex64"))

    with pytest.raises(RuntimeError, match="c.size must be same as x.size"):
        plan = Plan(1, (8,), dtype="complex64")
        plan.setpts(np.ones(3, dtype="float32"))
        plan.execute(np.ones(4, dtype="complex64"))

    with pytest.raises(RuntimeError, match="c.ndim must be 2 if n_trans > 1"):
        plan = Plan(1, (8,), n_trans=2, dtype="complex64")
        plan.setpts(np.ones(3, dtype="float32"))
        plan.execute(np.ones(2, dtype="complex64"))

    with pytest.raises(RuntimeError, match=r"c.shape must be \(n_trans, x.size\)"):
        plan = Plan(1, (8,), n_trans=2, dtype="complex64")
        plan.setpts(np.ones(3, dtype="float32"))
        plan.execute(np.ones((2, 4), dtype="complex64"))

    with pytest.raises(RuntimeError, match="same as the problem dimension"):
        plan = Plan(2, (8,), dtype="complex64")
        plan.setpts(np.ones(3, dtype="float32"))
        plan.execute(np.ones((1, 2, 8), dtype="complex64"))

    with pytest.raises(RuntimeError, match=r"same as the problem dimension \+ 1 for"):
        plan = Plan(2, (8,), n_trans=2, dtype="complex64")
        plan.setpts(np.ones(3, dtype="float32"))
        plan.execute(np.ones(8, dtype="complex64"))

    with pytest.raises(RuntimeError, match=r"f\.shape\[0\] must be n_trans"):
        plan = Plan(2, (8,), n_trans=2, dtype="complex64")
        plan.setpts(np.ones(3, dtype="float32"))
        plan.execute(np.ones((3, 8), dtype="complex64"))

    with pytest.raises(RuntimeError, match=r"f\.shape is not consistent"):
        plan = Plan(2, (8,), dtype="complex64")
        plan.setpts(np.ones(3, dtype="float32"))
        plan.execute(np.ones(2, dtype="complex64"))

    with pytest.raises(RuntimeError, match=r"f\.shape is not consistent"):
        plan = Plan(2, (8, 9), dtype="complex64")
        plan.setpts(*np.ones((2, 3), dtype="float32"))
        plan.execute(np.ones((2, 9), dtype="complex64"))

    with pytest.raises(RuntimeError, match=r"f\.shape is not consistent"):
        plan = Plan(2, (8, 9, 10), dtype="complex64")
        plan.setpts(*np.ones((3, 3), dtype="float32"))
        plan.execute(np.ones((2, 9, 10), dtype="complex64"))

    with pytest.raises(RuntimeError, match=r"f\.ndim must be 1 or 2"):
        plan = Plan(3, 1, dtype="complex64")
        plan.setpts(np.ones(3, dtype="float32"), s=np.ones(3, dtype="float32"))
        plan.execute(np.ones(3, dtype="complex64"), out=np.ones((1, 2, 3), dtype="complex64"))

    with pytest.raises(RuntimeError, match=r"f\.size of must be nk"):
        plan = Plan(3, 1, dtype="complex64")
        plan.setpts(np.ones(3, dtype="float32"), s=np.ones(3, dtype="float32"))
        plan.execute(np.ones(3, dtype="complex64"), out=np.ones(4, dtype="complex64"))

    with pytest.raises(RuntimeError, match=r"f\.ndim must be 2"):
        plan = Plan(3, 1, n_trans=2, dtype="complex64")
        plan.setpts(np.ones(3, dtype="float32"), s=np.ones(3, dtype="float32"))
        plan.execute(np.ones((2, 3), dtype="complex64"), out=np.ones(3, dtype="complex64"))

    with pytest.raises(RuntimeError, match=r"f\.shape must be \(n_trans, nk\)"):
        plan = Plan(3, 1, n_trans=2, dtype="complex64")
        plan.setpts(np.ones(3, dtype="float32"), s=np.ones(3, dtype="float32"))
        plan.execute(np.ones((2, 3), dtype="complex64"), out=np.ones((2, 4), dtype="complex64"))

    with pytest.raises(RuntimeError, match="transform type invalid"):
        plan = Plan(4, (8,))



#!/usr/bin/env python
"""
This file contains low level python bindings for the cufinufft CUDA libraries.
Seperate bindings are provided for single and double precision libraries,
differentiated by 'f' suffix.
"""

import ctypes
import os
import warnings
import importlib.util
import pathlib
import numpy as np
from ctypes.util import find_library

from packaging.version import Version

from ctypes import c_double
from ctypes import c_int
from ctypes import c_int64
from ctypes import c_float
from ctypes import c_void_p

c_int64_p = ctypes.POINTER(c_int64)
c_float_p = ctypes.POINTER(c_float)
c_double_p = ctypes.POINTER(c_double)


# numpy.distutils has a bug that changes the logging level from under us. As a
# workaround, we save it and reset it later. This only happens on older
# versions of NumPy, so let's check the version before doing this.
reset_log_level = Version(np.__version__) < Version("1.25")

if reset_log_level:
    import logging
    log_level = logging.root.level

lib = None
# Try to load the library as installed in the Python package.
path = pathlib.Path(__file__).parent.resolve()
library_names = ["libcufinufft", "cufinufft"]
for lib_name in library_names:
    try:
        lib = np.ctypeslib.load_library(lib_name, path)
        break
    except OSError:
        # Paranoid, in case lib is set to something and then an exception is thrown
        lib = None

if reset_log_level:
    logging.root.setLevel(log_level)

if lib is None:
    # If that fails, try to load the library from the system path.
    libname = find_library('cufinufft')
    if libname is not None:
        lib = ctypes.cdll.LoadLibrary(libname)
        # we probably should add a version check and trow a warning if the version is different
    else:
        # if that does not work, cufinufft is not installed correctly.
        raise ImportError("Failed to find a suitable cufinufft library.")


def _get_NufftOpts():
    fields = [
        ('upsampfac', c_double),
        ('gpu_method', c_int),
        ('gpu_sort', c_int),
        ('gpu_binsizex', c_int),
        ('gpu_binsizey', c_int),
        ('gpu_binsizez', c_int),
        ('gpu_obinsizex', c_int),
        ('gpu_obinsizey', c_int),
        ('gpu_obinsizez', c_int),
        ('gpu_maxsubprobsize', c_int),
        ('gpu_kerevalmeth', c_int),
        ('gpu_spreadinterponly', c_int),
        ('gpu_maxbatchsize', c_int),
        ('gpu_device_id', c_int),
        ('gpu_stream', c_void_p),
        ('modeord', c_int)
    ]
    return fields


class NufftOpts(ctypes.Structure):
    pass


NufftOpts._fields_ = _get_NufftOpts()


CufinufftPlan = c_void_p
CufinufftPlanf = c_void_p

CufinufftPlan_p = ctypes.POINTER(CufinufftPlan)
CufinufftPlanf_p = ctypes.POINTER(CufinufftPlanf)

NufftOpts_p = ctypes.POINTER(NufftOpts)

_default_opts = lib.cufinufft_default_opts
_default_opts.argtypes = [NufftOpts_p]
_default_opts.restype = None

_make_plan = lib.cufinufft_makeplan
_make_plan.argtypes = [
    c_int, c_int, c_int64_p, c_int,
    c_int, c_double, CufinufftPlan_p, NufftOpts_p]
_make_plan.restypes = c_int

_make_planf = lib.cufinufftf_makeplan
_make_planf.argtypes = [
    c_int, c_int, c_int64_p, c_int,
    c_int, c_float, CufinufftPlanf_p, NufftOpts_p]
_make_planf.restypes = c_int

_set_pts = lib.cufinufft_setpts
_set_pts.argtypes = [
    c_void_p, c_int64, c_void_p, c_void_p, c_void_p, ctypes.c_int, c_double_p,
    c_double_p, c_double_p]
_set_pts.restype = c_int

_set_ptsf = lib.cufinufftf_setpts
_set_ptsf.argtypes = [
    c_void_p, c_int64, c_void_p, c_void_p, c_void_p, ctypes.c_int, c_float_p,
    c_float_p, c_float_p]
_set_ptsf.restype = c_int

_exec_plan = lib.cufinufft_execute
_exec_plan.argtypes = [c_void_p, c_void_p, c_void_p]
_exec_plan.restype = c_int

_exec_planf = lib.cufinufftf_execute
_exec_planf.argtypes = [c_void_p, c_void_p, c_void_p]
_exec_planf.restype = c_int

_destroy_plan = lib.cufinufft_destroy
_destroy_plan.argtypes = [c_void_p]
_destroy_plan.restype = c_int

_destroy_planf = lib.cufinufftf_destroy
_destroy_planf.argtypes = [c_void_p]
_destroy_planf.restype = c_int



from cufinufft import Plan, _compat

def nufft1d1(x, data, n_modes=None, out=None, eps=1e-6, isign=1, **kwargs):
    return _invoke_plan(1, 1, x, None, None, data, out, isign, eps, n_modes,
            kwargs)

def nufft1d2(x, data, out=None, eps=1e-6, isign=-1, **kwargs):
    return _invoke_plan(1, 2, x, None, None, data, out, isign, eps, None,
            kwargs)

def nufft2d1(x, y, data, n_modes=None, out=None, eps=1e-6, isign=1, **kwargs):
    return _invoke_plan(2, 1, x, y, None, data, out, isign, eps, n_modes,
            kwargs)

def nufft2d2(x, y, data, out=None, eps=1e-6, isign=-1, **kwargs):
    return _invoke_plan(2, 2, x, y, None, data, out, isign, eps, None, kwargs)

def nufft3d1(x, y, z, data, n_modes=None, out=None, eps=1e-6, isign=1,
        **kwargs):
    return _invoke_plan(3, 1, x, y, z, data, out, isign, eps, n_modes, kwargs)

def nufft3d2(x, y, z, data, out=None, eps=1e-6, isign=-1, **kwargs):
    return _invoke_plan(3, 2, x, y, z, data, out, isign, eps, None, kwargs)

def _invoke_plan(dim, nufft_type, x, y, z, data, out, isign, eps,
        n_modes=None, kwargs=None):
    dtype = _compat.get_array_dtype(data)

    n_trans = _get_ntrans(dim, nufft_type, data)

    if nufft_type == 1 and out is not None:
        n_modes = out.shape[-dim:]
    if nufft_type == 2:
        n_modes = data.shape[-dim:]

    plan = Plan(nufft_type, n_modes, n_trans, eps, isign, dtype, **kwargs)

    plan.setpts(x, y, z)

    if out is None:
        out = plan.execute(data)
    else:
        plan.execute(data, out=out)

    return out


def _get_ntrans(dim, nufft_type, data):
    if nufft_type == 1:
        expect_dim = 1
    else:
        expect_dim = dim

    if data.ndim == expect_dim:
        n_trans = 1
    else:
        n_trans = data.shape[0]

    return n_trans


def _wrap_docstring(docstring, tw=80, min_spacing=2):
    lines = docstring.expandtabs().splitlines()

    for k, line in enumerate(lines):
        if len(line) > tw:
            last_space = line[:tw].rfind(' ')
            indent_level = line.rfind(' ' * min_spacing) + min_spacing

            lines[k] = line[:last_space]

            new_line = (' ' * indent_level) + line[last_space + 1:]

            # Check if the indentation level continues on next line. If so,
            # concatenate, otherwise insert new line.
            if len(lines[k + 1]) - len(lines[k + 1].lstrip()) >= indent_level:
                lines[k + 1] = new_line + ' ' + lines[k + 1].lstrip()
            else:
                lines.insert(k + 1, new_line)

    docstring = '\n'.join(lines)

    return docstring


def _set_nufft_doc(f, dim, tp):
    doc_nufft1 = \
    """{dim}D type-1 (nonuniform to uniform) complex NUFFT

    ::

      {pt_spacing}        M-1
      f[{pt_idx}] = SUM c[j] exp(+/-i {pt_inner})
      {pt_spacing}        j=0

          for {pt_constraint}

    Args:
{pts_doc}
      c         (complex[M] or complex[n_tr, M]): source strengths.
      n_modes   (integer or integer tuple of length {dim}, optional): number of
                uniform Fourier modes requested {modes_tuple}. May be even or odd; in
                either case, modes {pt_idx} are integers satisfying {pt_constraint}.
                Must be specified if ``out`` is not given.
      out       (complex[{modes}] or complex[n_tr, {modes}], optional): output array
                for Fourier mode values. If ``n_modes`` is specifed, the shape
                must match, otherwise ``n_modes`` is inferred from ``out``.
      eps       (float, optional): precision requested (>1e-16).
      isign     (int, optional): if non-negative, uses positive sign in
                exponential, otherwise negative sign.
      **kwargs  (optional): other options may be specified, see the ``Plan``
                constructor for details.

    .. note::

      The output is written into the ``out`` array if supplied.

    Returns:
      complex[{modes}] or complex[n_tr, {modes}]: The resulting array.

    Example (CuPy):
    ::

      import cupy as cp
      import cufinufft

      # number of nonuniform points
      M = 100

      # the nonuniform points
{pts_generate}

      # their complex strengths
      c = (cp.random.standard_normal(size=M)
           + 1J * cp.random.standard_normal(size=M))

      # desired number of Fourier modes
      {modes} = {sample_modes}

      # calculate the type-1 NUFFT
      f = cufinufft.nufft{dim}d1({pts}, c, {modes_tuple})
    """

    doc_nufft2 = \
    """{dim}D type-2 (uniform to nonuniform) complex NUFFT

    ::

      c[j] = SUM f[{pt_idx}] exp(+/-i {pt_inner})
             {pt_idx}

          for j = 0, ..., M-1, where the sum is over {pt_constraint}

    Args:
{pts_doc}
      f         (complex[{modes}] or complex[n_tr, {modes}]): Fourier mode
                coefficients, where {modes} may be even or odd. In either case
                the mode indices {pt_idx} satisfy {pt_constraint}.
      out       (complex[M] or complex[n_tr, M], optional): output array
                at targets.
      eps       (float, optional): precision requested (>1e-16).
      isign     (int, optional): if non-negative, uses positive sign in
                exponential, otherwise negative sign.
      **kwargs  (optional): other options may be specified, see the ``Plan``
                constructor for details.

    .. note::

      The output is written into the ``out`` array if supplied.

    Returns:
      complex[M] or complex[n_tr, M]: The resulting array.

    Example (CuPy):
    ::

      import cupy as cp
      import cufinufft

      # number of nonuniform points
      M = 100

      # the nonuniform points
{pts_generate}

      # number of Fourier modes
      {modes} = {sample_modes}

      # the Fourier mode coefficients
      f = (cp.random.standard_normal(size={modes_tuple})
           + 1J * cp.random.standard_normal(size={modes_tuple}))

      # calculate the type-2 NUFFT
      c = cufinufft.nufft{dim}d2({pts}, f)
    """

    doc_nufft = {1: doc_nufft1, 2: doc_nufft2}

    pts = ('x', 'y', 'z')
    sample_modes = (50, 75, 100)

    dims = range(1, dim + 1)

    v = {}

    v['dim'] = dim

    v['modes'] = ', '.join('N{}'.format(i) for i in dims)
    v['modes_tuple'] = '(' + v['modes'] + (', ' if dim == 1 else '') + ')'
    v['pt_idx'] = ', '.join('k{}'.format(i) for i in dims)
    v['pt_spacing'] = ' ' * (len(v['pt_idx']) - 2)
    v['pt_inner'] = ' + '.join('k{0} {1}(j)'.format(i, x) for i, x in zip(dims, pts[:dim]))
    v['pt_constraint'] = ', '.join('-N{0}/2 <= k{0} <= (N{0}-1)/2'.format(i) for i in dims)
    v['pts_doc'] = '\n'.join('      {}         (float[M]): nonuniform points, in the interval [-pi, pi), values outside will be folded'.format(x) for x in pts[:dim])

    # for example
    v['pts'] = ', '.join(str(x) for x in pts[:dim])
    v['pts_generate'] = '\n'.join('      {} = 2 * cp.pi * cp.random.uniform(size=M)'.format(x) for x in pts[:dim])
    v['sample_modes'] = ', '.join(str(n) for n in sample_modes[:dim])

    if dim > 1:
        v['pt_inner'] = '(' + v['pt_inner'] + ')'

    f.__doc__ = _wrap_docstring(doc_nufft[tp].format(**v))


_set_nufft_doc(nufft1d1, 1, 1)
_set_nufft_doc(nufft1d2, 1, 2)
_set_nufft_doc(nufft2d1, 2, 1)
_set_nufft_doc(nufft2d2, 2, 2)
_set_nufft_doc(nufft3d1, 3, 1)
_set_nufft_doc(nufft3d2, 3, 2)



# ahb updated for v2.0 interface, and complex c. Bug seems to have been fixed :)

import numpy as np

from finufft import nufft2d1

c = np.complex128(np.random.rand(2))

omega = np.arange(4).reshape((2, 2)) / 3 * np.pi

x0 = omega[:, 0]
y0 = omega[:, 1]

f0 = np.zeros((4, 4), order='F', dtype=np.complex128)

nufft2d1(x0, y0, c, f0.shape, out=f0, eps=1e-14)

x1 = x0.copy()
y1 = y0.copy()

f1 = np.zeros((4, 4), order='F', dtype=np.complex128)

nufft2d1(x1, y1, c, f1.shape, out=f1, eps=1e-14)

print('difference: %e' % np.linalg.norm(f0 - f1))



import cufinufft

import torch

# number of nonuniform points
M = 100000

# grid size
N = 200000

# generate positions for the nonuniform points and the coefficients
x_gpu = 2 * torch.pi * torch.rand(size=(M,)).cuda()
c_gpu = (torch.randn(size=(M,)) + 1J * torch.randn(size=(M,))).cuda()

# compute the transform
f_gpu = cufinufft.nufft1d1(x_gpu, c_gpu, (N,))

# move results off the GPU
f = f_gpu.cpu()



"""
Discover and run Python example scripts as unit tests.
"""

import os
import subprocess
import sys
from pathlib import Path

import pytest

examples_dir = os.path.join(Path(__file__).resolve().parents[1], "examples")

scripts = []
for filename in os.listdir(examples_dir):
    if filename.endswith(".py"):
        scripts.append(os.path.join(examples_dir, filename))

@pytest.mark.parametrize("filename", scripts)
def test_example(filename, request):
    # Extract framework from format `example_framework.py`.
    framework = Path(filename).stem.split("_")[-1]

    if framework in request.config.getoption("framework"):
        subprocess.check_call([sys.executable, filename])
    else:
        pytest.skip("Example not in list of frameworks")



import pytest

import numpy as np
from ctypes.util import find_library

@pytest.mark.skip(reason="Patching seems to fail in CI")
def test_fallback(mocker):
    def fake_load_library(lib_name, path):
        if lib_name in ["libfinufft", "finufft"]:
            raise OSError()
        else:
            return np.ctypeslib.load_library(lib_name, path)

    mocker.patch("numpy.ctypeslib.load_library", fake_load_library)

    if find_library("finufft") is None:
        with pytest.raises(ImportError, match="suitable finufft"):
            import finufft
    else:
        import finufft



import pytest

import numpy as np

import cufinufft
from cufinufft import _compat

import utils

DTYPES = [np.float32, np.float64]
SHAPES = [(16,), (16, 16), (16, 16, 16)]
N_TRANS = [(), (1,), (2,)]
MS = [256, 1024, 4096]
TOLS = [1e-3, 1e-6]
OUTPUT_ARGS = [False, True]

@pytest.mark.parametrize("dtype", DTYPES)
@pytest.mark.parametrize("n_trans", N_TRANS)
@pytest.mark.parametrize("shape", SHAPES)
@pytest.mark.parametrize("M", MS)
@pytest.mark.parametrize("tol", TOLS)
@pytest.mark.parametrize("output_arg", OUTPUT_ARGS)
def test_simple_type1(to_gpu, to_cpu, dtype, shape, n_trans, M, tol, output_arg):
    real_dtype = dtype
    complex_dtype = utils._complex_dtype(dtype)

    dim = len(shape)

    # Select which function to call based on dimension.
    fun = {1: cufinufft.nufft1d1,
           2: cufinufft.nufft2d1,
           3: cufinufft.nufft3d1}[dim]

    k, c = utils.type1_problem(dtype, shape, M, n_trans=n_trans)

    k_gpu = to_gpu(k)
    c_gpu = to_gpu(c)

    if output_arg:
        # Ensure that output array has proper shape i.e., (N1, ...) for no
        # batch, (1, N1, ...) for batch of size one, and (n, N1, ...) for
        # batch of size n.
        fk_gpu = _compat.array_empty_like(c_gpu, n_trans + shape,
                dtype=complex_dtype)

        fun(*k_gpu, c_gpu, out=fk_gpu, eps=tol)
    else:
        fk_gpu = fun(*k_gpu, c_gpu, shape, eps=tol)

    fk = to_cpu(fk_gpu)

    utils.verify_type1(k, c, fk, tol)


@pytest.mark.parametrize("dtype", DTYPES)
@pytest.mark.parametrize("shape", SHAPES)
@pytest.mark.parametrize("n_trans", N_TRANS)
@pytest.mark.parametrize("M", MS)
@pytest.mark.parametrize("tol", TOLS)
@pytest.mark.parametrize("output_arg", OUTPUT_ARGS)
def test_simple_type2(to_gpu, to_cpu, dtype, shape, n_trans, M, tol, output_arg):
    real_dtype = dtype
    complex_dtype = utils._complex_dtype(dtype)

    dim = len(shape)

    fun = {1: cufinufft.nufft1d2,
           2: cufinufft.nufft2d2,
           3: cufinufft.nufft3d2}[dim]

    k, fk = utils.type2_problem(dtype, shape, M, n_trans=n_trans)

    k_gpu = to_gpu(k)
    fk_gpu = to_gpu(fk)

    if output_arg:
        c_gpu = _compat.array_empty_like(fk_gpu, n_trans + (M,),
                dtype=complex_dtype)

        fun(*k_gpu, fk_gpu, eps=tol, out=c_gpu)
    else:
        c_gpu = fun(*k_gpu, fk_gpu, eps=tol)

    c = to_cpu(c_gpu)

    utils.verify_type2(k, fk, c, tol)



import pytest

import utils


def pytest_addoption(parser):
    parser.addoption("--framework", action="append", default=[], help="List of frameworks")

def pytest_generate_tests(metafunc):
    if "framework" in metafunc.fixturenames:
        metafunc.parametrize("framework", metafunc.config.getoption("framework"))

@pytest.fixture
def to_gpu(framework):
    to_gpu, _ = utils.transfer_funcs(framework)

    return to_gpu


@pytest.fixture
def to_cpu(framework):
    _, to_cpu = utils.transfer_funcs(framework)

    return to_cpu



"""
Demonstrate the type 2 NUFFT using cuFINUFFT
"""

import numpy as np

import pycuda.autoinit
from pycuda.gpuarray import to_gpu

import cufinufft

# Set up parameters for problem.
N1, N2 = 37, 41                 # Size of uniform grid
M = 17                          # Number of nonuniform points
n_transf = 2                    # Number of input arrays
eps = 1e-6                      # Requested tolerance
dtype = np.float32              # Datatype (real)
complex_dtype = np.complex64    # Datatype (complex)

# Generate coordinates of non-uniform points.
x = np.random.uniform(-np.pi, np.pi, size=M)
y = np.random.uniform(-np.pi, np.pi, size=M)

# Generate grid values.
fk = (np.random.standard_normal((n_transf, N1, N2))
      + 1j * np.random.standard_normal((n_transf, N1, N2)))

# Cast to desired datatype.
x = x.astype(dtype)
y = y.astype(dtype)
fk = fk.astype(complex_dtype)

# Initialize the plan and set the points.
plan = cufinufft.Plan(2, (N1, N2), n_transf, eps=eps, dtype=complex_dtype)
plan.setpts(to_gpu(x), to_gpu(y))

# Execute the plan, reading from the uniform grid fk and storing the result
# in c_gpu.
c_gpu = plan.execute(to_gpu(fk))

# Retreive the result from the GPU.
c = c_gpu.get()

# Check accuracy of the transform at index jt.
jt = M // 2

for i in range(n_transf):
    # Calculate the true value of the type 2 transform at the index jt.
    m, n = np.mgrid[-(N1 // 2):(N1 + 1) // 2, -(N2 // 2):(N2 + 1) // 2]
    c_true = np.sum(fk[i] * np.exp(-1j * (m * x[jt] + n * y[jt])))

    # Calculate the absolute and relative error.
    err = np.abs(c[i, jt] - c_true)
    rel_err = err / np.max(np.abs(c[i]))

    print(f"[{i}] Absolute error on point [{jt}] is {err:.3g}")
    print(f"[{i}] Relative error on point [{jt}] is {rel_err:.3g}")

    assert(rel_err < 15 * eps)



# demo of vectorized 2D type 1 FINUFFT in python. Should stay close to docs/python.rst
# Barnett 8/19/20

import numpy as np
import finufft
import time
np.random.seed(42)

# number of nonuniform points
M = 100000

# the nonuniform points in the square [0,2pi)^2
x = 2 * np.pi * np.random.uniform(size=M)
y = 2 * np.pi * np.random.uniform(size=M)

# number of transforms
K = 4

# generate K stacked strength arrays
c = (np.random.standard_normal(size=(K, M))
     + 1J * np.random.standard_normal(size=(K, M)))

# desired number of Fourier modes (in x,y directions respectively)
N1 = 1000
N2 = 2000

# calculate the K transforms simultaneously (K is inferred from c.shape)
t0 = time.time()
f = finufft.nufft2d1(x, y, c, (N1,N2), eps=1e-9)
print("vectorized finufft2d1 done in {0:.2g} s.".format(time.time()-t0))
print(f.shape)

k1 = 376     # do a math check, for a single output mode index (k1,k2)
k2 = -1000
t = K-1      # from the t'th transform
assert((k1>=-N1/2.) & (k1<N1/2.))   # float division easier here
assert((k2>=-N2/2.) & (k2<N2/2.))
assert((t>=0) & (t<K))
ftest = sum(c[t,:] * np.exp(1.j*(k1*x + k2*y)))
err = np.abs(f[t, k1+N1//2, k2+N2//2] - ftest) / np.max(np.abs(f))
print("Error relative to max: {0:.2e}".format(err))



# Jeremy Magland, Sept 2017.
# Alex Barnett fixed integer division issue in python v2 vs v3
#              (affected 1/2, 1/3), 10/13/17.
# Removed ms etc from ?d2 interfaces, 10/31/17. Less reruns 2/14/18.
# 2d1many and 2d2many added, painfully, Barnett 7/29/18

import numpy as np
import finufft
import math
import time

def compute_error(Xest,Xtrue):
	numer=np.sqrt(np.sum(np.abs((Xest-Xtrue))**2));
	denom=np.sqrt(np.sum(np.abs(Xtrue)**2));    # rel l2 norm
	if (denom!=0):
		return numer/denom
	else:
		return 0

def print_report(label,elapsed,Xest,Xtrue,npts):
	print(label+':')
	print('    Est rel l2 err  %.3g' % (compute_error(Xest,Xtrue)))
	print('    CPU time (sec)  %.3g' % (elapsed))
	if (elapsed>0):
		print('    tot NU pts/sec  %.3g' % (npts/elapsed))
	print('')

def accuracy_speed_tests(num_nonuniform_points,num_uniform_points,eps):
	nj,nk = int(num_nonuniform_points),int(num_nonuniform_points)
	iflag=1
	num_samples=int(np.minimum(5,num_uniform_points*0.5+1)) # number of outputs used for estimating accuracy; is small for speed

	print('Accuracy and speed tests for %d nonuniform points and eps=%g (error estimates use %d samples per run)' % (num_nonuniform_points,eps,num_samples))

	# for doing the error estimates
	Xest=np.zeros(num_samples,dtype=np.complex128)
	Xtrue=np.zeros(num_samples,dtype=np.complex128)

	###### 1-d cases ........................................................
	ms=int(num_uniform_points)

	xj=np.random.rand(nj)*2*math.pi-math.pi
	cj=np.random.rand(nj)+1j*np.random.rand(nj);
	fk=np.zeros([ms],dtype=np.complex128)
	timer=time.time()
	ret=finufft.nufft1d1(xj,cj,ms,fk,eps,iflag)
	elapsed=time.time()-timer

	k=np.arange(-np.floor(ms/2),np.floor((ms-1)/2+1))
	for ii in np.arange(0,num_samples):
		Xest[ii]=np.sum(cj * np.exp(1j*k[ii]*xj))
		Xtrue[ii]=fk[ii]
	print_report('finufft1d1',elapsed,Xest,Xtrue,nj)

	xj=np.random.rand(nj)*2*math.pi-math.pi
	cj=np.zeros([nj],dtype=np.complex128);
	fk=np.random.rand(ms)+1j*np.random.rand(ms);
	timer=time.time()
	ret=finufft.nufft1d2(xj,fk,cj,eps,iflag)
	elapsed=time.time()-timer

	k=np.arange(-np.floor(ms/2),np.floor((ms-1)/2+1))
	for ii in np.arange(0,num_samples):
		Xest[ii]=np.sum(fk * np.exp(1j*k*xj[ii]))
		Xtrue[ii]=cj[ii]
	print_report('finufft1d2',elapsed,Xest,Xtrue,nj)

	x=np.random.rand(nj)*2*math.pi-math.pi
	c=np.random.rand(nj)+1j*np.random.rand(nj);
	s=np.random.rand(nk)*2*math.pi-math.pi
	f=np.zeros([nk],dtype=np.complex128)
	timer=time.time()
	ret=finufft.nufft1d3(x,c,s,f,eps,iflag)
	elapsed=time.time()-timer

	for ii in np.arange(0,num_samples):
		Xest[ii]=np.sum(c * np.exp(1j*s[ii]*x))
		Xtrue[ii]=f[ii]
	print_report('finufft1d3',elapsed,Xest,Xtrue,nj+nk)

	###### 2-d cases ....................................................
	ms=int(np.ceil(np.sqrt(num_uniform_points)))
	mt=ms

	xj=np.random.rand(nj)*2*math.pi-math.pi
	yj=np.random.rand(nj)*2*math.pi-math.pi
	cj=np.random.rand(nj)+1j*np.random.rand(nj)
	fk=np.zeros([ms,mt],dtype=np.complex128)
	timer=time.time()
	ret=finufft.nufft2d1(xj,yj,cj,(ms,mt),fk,eps,iflag)
	elapsed=time.time()-timer

	Ks,Kt=np.mgrid[-np.floor(ms/2):np.floor((ms-1)/2+1),-np.floor(mt/2):np.floor((mt-1)/2+1)]

	for ii in np.arange(0,num_samples):
		Xest[ii]=np.sum(cj * np.exp(1j*(Ks.ravel()[ii]*xj+Kt.ravel()[ii]*yj)))
		Xtrue[ii]=fk.ravel()[ii]
	print_report('finufft2d1',elapsed,Xest,Xtrue,nj)

	## 2d1many:
	ndata = 8       # how many vectors to do
	cj=np.array(np.random.rand(ndata,nj)+1j*np.random.rand(ndata,nj))
	fk=np.zeros([ndata,ms,mt],dtype=np.complex128)
	timer=time.time()
	ret=finufft.nufft2d1(xj,yj,cj,ms,fk,eps,iflag)
	elapsed=time.time()-timer

	dtest = ndata-1    # which of the ndata to test (in 0,..,ndata-1)
	for ii in np.arange(0,num_samples):
		Xest[ii]=np.sum(cj[dtest,:] * np.exp(1j*(Ks.ravel()[ii]*xj+Kt.ravel()[ii]*yj)))   # note fortran-ravel-order needed throughout - mess.
		Xtrue[ii]=fk.ravel()[ii + dtest*ms*mt]       # hack the offset in fk array - has to be better way
	print_report('finufft2d1many',elapsed,Xest,Xtrue,ndata*nj)

	# 2d2
	xj=np.random.rand(nj)*2*math.pi-math.pi
	yj=np.random.rand(nj)*2*math.pi-math.pi
	cj=np.zeros([nj],dtype=np.complex128);
	fk=np.random.rand(ms,mt)+1j*np.random.rand(ms,mt);
	timer=time.time()
	ret=finufft.nufft2d2(xj,yj,fk,cj,eps,iflag)
	elapsed=time.time()-timer

	Ks,Kt=np.mgrid[-np.floor(ms/2):np.floor((ms-1)/2+1),-np.floor(mt/2):np.floor((mt-1)/2+1)]
	for ii in np.arange(0,num_samples):
		Xest[ii]=np.sum(fk * np.exp(1j*(Ks*xj[ii]+Kt*yj[ii])))
		Xtrue[ii]=cj[ii]
	print_report('finufft2d2',elapsed,Xest,Xtrue,nj)

	# 2d2many (using same ndata and dtest as 2d1many; see above)
	cj=np.zeros([ndata,nj],dtype=np.complex128);
	fk=np.array(np.random.rand(ndata,ms,mt)+1j*np.random.rand(ndata,ms,mt))
	timer=time.time()
	ret=finufft.nufft2d2(xj,yj,fk,cj,eps,iflag)
	elapsed=time.time()-timer

	for ii in np.arange(0,num_samples):
		Xest[ii]=np.sum(fk[dtest,:,:] * np.exp(1j*(Ks*xj[ii]+Kt*yj[ii])))
		Xtrue[ii]=cj[dtest,ii]
	print_report('finufft2d2many',elapsed,Xest,Xtrue,ndata*nj)
	
	# 2d3
	x=np.random.rand(nj)*2*math.pi-math.pi
	y=np.random.rand(nj)*2*math.pi-math.pi
	c=np.random.rand(nj)+1j*np.random.rand(nj);
	s=np.random.rand(nk)*2*math.pi-math.pi
	t=np.random.rand(nk)*2*math.pi-math.pi
	f=np.zeros([nk],dtype=np.complex128)
	timer=time.time()
	ret=finufft.nufft2d3(x,y,c,s,t,f,eps,iflag)
	elapsed=time.time()-timer

	for ii in np.arange(0,num_samples):
		Xest[ii]=np.sum(c * np.exp(1j*(s[ii]*x+t[ii]*y)))
		Xtrue[ii]=f[ii]
	print_report('finufft2d3',elapsed,Xest,Xtrue,nj+nk)

	###### 3-d cases ............................................................
	ms=int(np.ceil(num_uniform_points**(1.0/3)))
	mt=ms
	mu=ms

	xj=np.random.rand(nj)*2*math.pi-math.pi
	yj=np.random.rand(nj)*2*math.pi-math.pi
	zj=np.random.rand(nj)*2*math.pi-math.pi
	cj=np.random.rand(nj)+1j*np.random.rand(nj);
	fk=np.zeros([ms,mt,mu],dtype=np.complex128)
	timer=time.time()
	ret=finufft.nufft3d1(xj,yj,zj,cj,fk.shape,fk,eps,iflag)
	elapsed=time.time()-timer

	Ks,Kt,Ku=np.mgrid[-np.floor(ms/2):np.floor((ms-1)/2+1),-np.floor(mt/2):np.floor((mt-1)/2+1),-np.floor(mu/2):np.floor((mu-1)/2+1)]
	for ii in np.arange(0,num_samples):
		Xest[ii]=np.sum(cj * np.exp(1j*(Ks.ravel()[ii]*xj+Kt.ravel()[ii]*yj+Ku.ravel()[ii]*zj)))
		Xtrue[ii]=fk.ravel()[ii]
	print_report('finufft3d1',elapsed,Xest,Xtrue,nj)

	xj=np.random.rand(nj)*2*math.pi-math.pi
	yj=np.random.rand(nj)*2*math.pi-math.pi
	zj=np.random.rand(nj)*2*math.pi-math.pi
	cj=np.zeros([nj],dtype=np.complex128);
	fk=np.random.rand(ms,mt,mu)+1j*np.random.rand(ms,mt,mu);
	timer=time.time()
	ret=finufft.nufft3d2(xj,yj,zj,fk,cj,eps,iflag)
	elapsed=time.time()-timer

	Ks,Kt,Ku=np.mgrid[-np.floor(ms/2):np.floor((ms-1)/2+1),-np.floor(mt/2):np.floor((mt-1)/2+1),-np.floor(mu/2):np.floor((mu-1)/2+1)]
	for ii in np.arange(0,num_samples):
		Xest[ii]=np.sum(fk * np.exp(1j*(Ks*xj[ii]+Kt*yj[ii]+Ku*zj[ii])))
		Xtrue[ii]=cj[ii]
	print_report('finufft3d2',elapsed,Xest,Xtrue,nj)

	x=np.random.rand(nj)*2*math.pi-math.pi
	y=np.random.rand(nj)*2*math.pi-math.pi
	z=np.random.rand(nj)*2*math.pi-math.pi
	c=np.random.rand(nj)+1j*np.random.rand(nj);
	s=np.random.rand(nk)*2*math.pi-math.pi
	t=np.random.rand(nk)*2*math.pi-math.pi
	u=np.random.rand(nk)*2*math.pi-math.pi
	f=np.zeros([nk],dtype=np.complex128)
	timer=time.time()
	ret=finufft.nufft3d3(x,y,z,c,s,t,u,f,eps,iflag)
	elapsed=time.time()-timer

	for ii in np.arange(0,num_samples):
		Xest[ii]=np.sum(c * np.exp(1j*(s[ii]*x+t[ii]*y+u[ii]*z)))
		Xtrue[ii]=f[ii]
	print_report('finufft3d3',elapsed,Xest,Xtrue,nj+nk)



"""
Demonstrate the type 1 NUFFT using cuFINUFFT
"""

import numpy as np

import pycuda.autoinit
from pycuda.gpuarray import to_gpu

import cufinufft

# Set up parameters for problem.
N1, N2 = 59, 61                 # Size of uniform grid
M = 100                         # Number of nonuniform points
n_transf = 2                    # Number of input arrays
eps = 1e-6                      # Requested tolerance
dtype = np.float32              # Datatype (real)
complex_dtype = np.complex64    # Datatype (complex)

# Generate coordinates of non-uniform points.
x = np.random.uniform(-np.pi, np.pi, size=M)
y = np.random.uniform(-np.pi, np.pi, size=M)

# Generate source strengths.
c = (np.random.standard_normal((n_transf, M))
     + 1j * np.random.standard_normal((n_transf, M)))

# Cast to desired datatype.
x = x.astype(dtype)
y = y.astype(dtype)
c = c.astype(complex_dtype)

# Initialize the plan and set the points.
plan = cufinufft.Plan(1, (N1, N2), n_transf, eps=eps, dtype=complex_dtype)
plan.setpts(to_gpu(x), to_gpu(y))

# Execute the plan, reading from the strengths array c and storing the
# result in fk_gpu.
fk_gpu = plan.execute(to_gpu(c))

# Retreive the result from the GPU.
fk = fk_gpu.get()

# Check accuracy of the transform at position (nt1, nt2).
nt1 = int(0.37 * N1)
nt2 = int(0.26 * N2)

for i in range(n_transf):
    # Calculate the true value of the type 1 transform at the uniform grid
    # point (nt1, nt2), which corresponds to the coordinate nt1 - N1 // 2 and
    # nt2 - N2 // 2.
    m, n = nt1 - N1 // 2, nt2 - N2 // 2
    fk_true = np.sum(c[i] * np.exp(1j * (m * x + n * y)))

    # Calculate the absolute and relative error.
    err = np.abs(fk[i, nt1, nt2] - fk_true)
    rel_err = err / np.max(np.abs(fk[i]))

    print(f"[{i}] Absolute error on mode [{nt1}, {nt2}] is {err:.3g}")
    print(f"[{i}] Relative error on mode [{nt1}, {nt2}] is {rel_err:.3g}")

    assert(rel_err < 10 * eps)



# demo of vectorized 2D type 1 FINUFFT in python via guru interface. Should stay close to docs/python.rst
# Lu 8/20/20

import numpy as np
import finufft
import time
np.random.seed(42)

# number of nonuniform points
M = 100000

# the nonuniform points in the square [0,2pi)^2
x = 2 * np.pi * np.random.uniform(size=M)
y = 2 * np.pi * np.random.uniform(size=M)

# number of transforms
K = 4

# generate K stacked strength arrays
c = (np.random.standard_normal(size=(K, M))
     + 1J * np.random.standard_normal(size=(K, M)))

# convert input data to single precision
x = x.astype('float32')
y = y.astype('float32')
c = c.astype('complex64')

# desired number of Fourier modes (in x,y directions respectively)
N1 = 1000
N2 = 2000

# specify type 1 transform
nufft_type = 1

# instantiate the plan (note n_trans must be set here), also setting tolerance:
t0 = time.time()
plan = finufft.Plan(nufft_type, (N1, N2), eps=1e-4, n_trans=K, dtype='float32')

# set the nonuniform points
plan.setpts(x, y)

# execute the plan (K transforms together, note c.shape must match)
f = plan.execute(c)
print("vectorized guru single-prec finufft2d1 done in {0:.2g} s.".format(time.time()-t0))

print(f.dtype)
print(f.shape)

k1 = 37     # do a math check, for a single output mode index (k1,k2)
k2 = -100
t = K-2      # from the t'th transform
assert((k1>=-N1/2.) & (k1<N1/2.))   # float division easier here
assert((k2>=-N2/2.) & (k2<N2/2.))
assert((t>=0) & (t<K))
ftest = sum(c[t,:] * np.exp(1.j*(k1*x + k2*y)))
err = np.abs(f[t, k1+N1//2, k2+N2//2] - ftest) / np.max(np.abs(f))
print("Error relative to max: {0:.2e}".format(err))



# demo of 1D type 1 FINUFFT in python. Should stay close to docs/python.rst
# Barnett 8/19/20

import numpy as np
import finufft
import time
np.random.seed(42)

# number of nonuniform points
M = 100000

# input nonuniform points
x = 2 * np.pi * np.random.uniform(size=M)

# their complex strengths
c = (np.random.standard_normal(size=M)
     + 1J * np.random.standard_normal(size=M))

# desired number of output Fourier modes
N = 1000000

# calculate the transform
t0 = time.time()
f = finufft.nufft1d1(x, c, N, eps=1e-9)
print("finufft1d1 done in {0:.2g} s.".format(time.time()-t0))

n = 142519   # do a math check, for a single output mode index n
assert((n>=-N/2.) & (n<N/2.))
ftest = sum(c * np.exp(1.j*n*x))
err = np.abs(f[n + N // 2] - ftest) / np.max(np.abs(f))
print("Error relative to max: {0:.2e}".format(err))



#!/usr/bin/env python3

import math
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import numpy as np
import re
import subprocess
import searchForTimeMetrics as stm

'''
A script to run a seris of finufftGuru_tests with varying parameters
Captures the stdout, parses it for timing statistics, and graphs speedup
ratio trends. 
'''


M_srcpts = 1e7 #1e6
tolerance = 1e-6
debug = 1

#modes array.
#first dimension counts are in modes[0,1,2]
#second dimension: modes[3,4,5]
#3rd dimension [6,7,8]
modes = [1e6,1,1,1e3,1e3,1,1e2,1e2,1e2]

dimensions = [1,2,3]
types = [1,2,3] 
#n_trials=[1,10,100]
n_trials = [1,10]


#data capture arrays
#Datapoints are speedup ratio := oldImplementationTime/newImplentationTime
totalTimeT1Ratio=[]
totalTimeT2Ratio=[]
totalTimeT3Ratio=[]
spreadT1Ratio=[]
spreadT2Ratio=[]
spreadT3Ratio=[]
fftwPlanT1Ratio=[]
fftwPlanT2Ratio=[]
fftwPlanT3Ratio=[]
fftT1Ratio=[]
fftT2Ratio=[]
fftT3Ratio=[]

#raw timing values

totalTimeT1_Old =[]
totalTimeT1_New =[]
totalTimeT2_Old =[]
totalTimeT2_New =[]
totalTimeT3_Old =[]
totalTimeT3_New =[]

spreadT1_Old =[]
spreadT1_New =[]
spreadT2_Old =[]
spreadT2_New =[]
spreadT3_Old =[]
spreadT3_New =[]

fftwPlanT1_Old =[]
fftwPlanT1_Old_initial =[]
fftwPlanT1_New =[]
fftwPlanT2_Old =[]
fftwPlanT2_Old_initial =[]
fftwPlanT2_New =[]
fftwPlanT3_Old =[]
fftwPlanT3_Old_initial =[]
fftwPlanT3_New =[]


fftT1_Old =[]
fftT1_New =[]
fftT2_Old =[]
fftT2_New =[]
fftT3_Old =[]
fftT3_New =[]

#do
for dim in dimensions:
    for ftype in types:
        for trial in n_trials:
            
            print( "./finufftGuru_test "+ str(trial)+ " " + str(ftype)+  " " +str(dim)+ " " + 
                                  str(modes[3*(dim-1)])+ " " + str(modes[3*(dim-1)+1])+ " " +  str(modes[3*(dim-1)+2])+ " " + 
                                  str(M_srcpts)+  " " + str(tolerance) + " " + str(debug)); 
            #execute the test for this set of parameters
            out =  subprocess.run(["./finufftGuru_test", str(trial), str(ftype), str(dim), 
                                  str(modes[3*(dim-1)]),str(modes[3*(dim-1)+1]),  str(modes[3*(dim-1)+2]), 
                                  str(M_srcpts), str(tolerance), str(debug)], 
                                 stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

            strOut = out.stdout.decode() #convert bytes to string
            print(strOut)

            
            #parse the output and syphon into data arrays

            ###############################################################################

            #gather total plan,setpts,execute,destroy
            planTime = stm.extractTime('(finufft_plan.*completed)(.*)',strOut)
            setPtsTime = stm.extractTime('(finufft_setpts.*completed)(.*)',strOut)
            execTime = stm.extractTime('(finufft_exec.*completed)(.*)', strOut)
            delTime = stm.extractTime('(finufft_destroy.*completed)(.*)',strOut)
            totalNewTime = round(planTime + setPtsTime + execTime + delTime,5)

            #gather old total time
            totalOldTime = stm.extractTime('(execute.*in)(.*)(or .*)' ,strOut)

            if(ftype == 1):
                totalTimeT1_New.append(totalNewTime)
                totalTimeT1_Old.append(totalOldTime)
            elif(ftype == 2):
                totalTimeT2_New.append(totalNewTime)
                totalTimeT2_Old.append(totalOldTime)
            else:
                totalTimeT3_New.append(totalNewTime)
                totalTimeT3_Old.append(totalOldTime)

            #total time speedup
            totalSpeedup = round(totalOldTime/totalNewTime,5)

            if(ftype == 1):
                totalTimeT1Ratio.append(totalSpeedup)
            elif(ftype == 2):
                totalTimeT2Ratio.append(totalSpeedup)
            else:
                totalTimeT3Ratio.append(totalSpeedup)

            ###############################################################################
                
            #spread (old) / [sort+spread]  (new)            
            newSort = stm.sumAllTime('(.*finufft_setpts.*sort)(.*)', strOut)

            #collect spreading if any
            newSpread = stm.extractTime('(.*finufft.*exec.*spread)(.*)' , strOut) 

            #collect interp if any
            newInterp = stm.extractTime('(.*finufft.*exec.*interp)(.*)',strOut)

            #collect the spread timings for each trial of old
            totalOldSpread = stm.sumAllTime('(.*spread.*ier)(.*)', strOut) #gets spread AND unspread (i.e. interpolation)

            spreadRatio = round(totalOldSpread/(newSort + newSpread + newInterp),5)

            if(ftype == 1):
                spreadT1Ratio.append(spreadRatio)
                spreadT1_New.append(round(newSort + newSpread + newInterp,5))
                spreadT1_Old.append(totalOldSpread)
            elif(ftype == 2):
                spreadT2Ratio.append(spreadRatio)
                spreadT2_New.append(round(newSort + newSpread + newInterp,5))
                spreadT2_Old.append(totalOldSpread)
            else:
                spreadT3Ratio.append(spreadRatio)
                spreadT3_New.append(round(newSort + newSpread + newInterp,5))
                spreadT3_Old.append(totalOldSpread)

            ###############################################################################

            #fftw_plan(old) / fftw_plan(new)
            planSciNotString = '(\(\d+\)[ \t]+)(\d*.?\d*e-\d* s)'
            planDecimalMatchString= '(\(\d+\)[ \t]+)(\d*\.?\d* s)'
            planWholeNumberMatchString= '(\(\d+\)[ \t]+)(\d+ s)' 

            #collect new fftw_plan time
            new_fftwPlan=0
            lineMatch = re.search("(.*make plan.*fftw plan \(\d+\).*)",strOut)
            if(lineMatch):
                fftwPlanVal = re.search(planSciNotString, lineMatch.group(0))
                if(not fftwPlanVal):
                    fftwPlanVal = re.search(planDecimalMatchString, lineMatch.group(0))
                if(not fftwPlanVal):
                    fftwPlanVal = re.search(wholeNumberMatchString, lineMatch.group(0))
                new_fftwPlan = float(fftwPlanVal.group(2).split('s')[0])  #strip off s
            new_fftwPlan = round(new_fftwPlan,5)    

            #collect the fftw_plan timings for each trial of old
            isInitial = True
            initialLookup = 0
            totalOldfftwPlan=0   
            lineMatch = re.findall('(?<!\[make plan\] )fftw plan \(64\).*', strOut) #all fftw plan lines that don't include "make plan" indicating old implm.
            if(lineMatch):
                for match in lineMatch:
                    oldfftwPlanVal = re.search(planSciNotString, match)
                    if(not oldfftwPlanVal):
                        oldfftwPlanVal = re.search(planDecimalMatchString, match)
                    if(not oldfftwPlanVal):
                        oldfftwPlanVal = re.search(planWholeNumberMatchString, match)
                    oldfftwPlanVal = float(oldfftwPlanVal.group(2).split('s')[0]) #trim off " s"
                    if(isInitial): #Capture the first fftwplan output - indicating initial construction time
                        initalLookup = oldfftwPlanVal
                        isInitial = False
                    
                    totalOldfftwPlan = totalOldfftwPlan + oldfftwPlanVal
            totalOldfftwPlan = round(totalOldfftwPlan,5)
            
            #These plan ratios include the initial old implementation plan construction!!
            fftwPlanRatio = round(totalOldfftwPlan/new_fftwPlan,5)
            
            if(ftype == 1):
                fftwPlanT1Ratio.append(fftwPlanRatio)
                fftwPlanT1_New.append(new_fftwPlan)
                fftwPlanT1_Old.append(totalOldfftwPlan)
                fftwPlanT1_Old_initial.append(initialLookup)
            elif(ftype == 2):
                fftwPlanT2Ratio.append(fftwPlanRatio)
                fftwPlanT2_New.append(new_fftwPlan)
                fftwPlanT2_Old.append(totalOldfftwPlan)
                fftwPlanT2_Old_initial.append(initialLookup)
            else:
                fftwPlanT3Ratio.append(fftwPlanRatio)
                fftwPlanT3_New.append(new_fftwPlan)
                fftwPlanT3_Old.append(totalOldfftwPlan)
                fftwPlanT3_Old_initial.append(initialLookup)
            
            ###############################################################################
            #fftw_exec(old) / fftw_exec(new)

            #collect new fft time
            new_fft = stm.extractTime("(.*finufft_exec.*fft)(.*)" , strOut)

            #collect the fftw_exec timings for each trial of old
            totalOldfft = stm.sumAllTime("(.*fft \(\d+ threads\))(.*)",strOut) 

            fftRatio = round(totalOldfft/new_fft,5)
            
            if(ftype == 1):
                fftT1Ratio.append(fftRatio)
                fftT1_New.append(new_fft)
                fftT1_Old.append(totalOldfft)
            elif(ftype == 2):
                fftT2Ratio.append(fftRatio)
                fftT2_New.append(new_fft)
                fftT2_Old.append(totalOldfft)
            else:
                fftT3Ratio.append(fftRatio)
                fftT3_New.append(new_fft)
                fftT3_Old.append(totalOldfft)

            
            ###############################################################################


#Construct the bar graph 
barWidth = 0.25
barDepth = 0.25

_t1ys = dimensions
_t2ys = [y + barWidth for y in _t1ys]
_t3ys = [y + barWidth for y in _t2ys]

_t1xs = n_trials 
_t2xs = [x + barDepth for x in _t1xs]
_t3xs = [x + barDepth for x in _t2xs]

_t1xx, _t1yy = np.meshgrid(_t1xs,_t1ys)
t1x,t1y = _t1xx.ravel(), _t1yy.ravel()

_t2xx, _t2yy = np.meshgrid(_t2xs,_t2ys)
t2x,t2y = _t2xx.ravel(), _t2yy.ravel()

_t3xx, _t3yy = np.meshgrid(_t3xs,_t3ys)
t3x,t3y = _t3xx.ravel(), _t3yy.ravel()

print("n_trials:" )
print(t1x)
print("dimensions: ")
print(t1y)
    
zbot = np.zeros(len(t1x))
widths = [barWidth]*len(t1x)
depths = [barDepth]*len(t1x)

fig = plt.figure()
#create legend
t1_proxy = plt.Rectangle((0,0),1,1,fc="r")
t2_proxy = plt.Rectangle((0,0),1,1,fc="b")
t3_proxy = plt.Rectangle((0,0),1,1,fc="g")


##################TotalSpeed BAR GRAPH####################################################
print("##################Total Time####################################################") 
print("\n")
print("Raw T1 Total Time New " + str(totalTimeT1_New))
print("Raw T1 Total Time Old " + str(totalTimeT1_Old))
print("TotalTime T1Ratio " + str(totalTimeT1Ratio))
print("\n")
print("Raw T2 Total Time New " + str(totalTimeT2_New))
print("Raw T2 Total Time Old " + str(totalTimeT2_Old))
print("TotalTime T2Ratio " + str(totalTimeT2Ratio))
print("\n")
print("Raw T3 Total Time New " + str(totalTimeT3_New))
print("Raw T3 Total Time Old " + str(totalTimeT3_Old))
print("TotalTime T3Ratio " + str(totalTimeT3Ratio))
print("\n")
ax1 = fig.add_subplot(221,projection='3d')

if(totalTimeT1Ratio):
    ax1.bar3d(t1x, t1y, zbot, widths, depths, totalTimeT1Ratio, shade=True, color='r', label='type1', alpha='1')
if(totalTimeT2Ratio):
    ax1.bar3d(t2x, t2y, zbot, widths, depths, totalTimeT2Ratio, shade=True, color='b', label='type2', alpha='1')
if(totalTimeT3Ratio):
    ax1.bar3d(t3x, t3y, zbot, widths, depths, totalTimeT3Ratio, shade=True, color='g', label='type3', alpha='1')

ax1.legend([t1_proxy,t2_proxy,t3_proxy], ['type1','type2','type3'])

plt.xlabel('n_trials')
plt.ylabel('Dimensions')
plt.yticks([y+barWidth+1 for y in range(len(t1y))], ['1', '2', '3'])
plt.title('totalOldTime/totalNewTime')


#### Speed Statistics SANS Initial Planning Time 

TotalSpeedRatioSansPlanT1 = (np.array(totalTimeT1_Old) - np.array(fftwPlanT1_Old_initial))/(np.array(totalTimeT1_New) - np.array(fftwPlanT1_New))
TotalSpeedRatioSansPlanT2 = (np.array(totalTimeT2_Old) - np.array(fftwPlanT2_Old_initial))/(np.array(totalTimeT2_New) - np.array(fftwPlanT2_New))
TotalSpeedRatioSansPlanT3 = (np.array(totalTimeT3_Old) - np.array(fftwPlanT3_Old_initial))/(np.array(totalTimeT3_New) - np.array(fftwPlanT3_New))


print("Total Speed Ratio SANS initial fftwPlan for old implm. and only fftwPlan for new")
print("T1: " + str(TotalSpeedRatioSansPlanT1))
print("T2: " + str(TotalSpeedRatioSansPlanT2))
print("T3: " + str(TotalSpeedRatioSansPlanT3))






##################SPREADING BAR GRAPH####################################################
print("##################SPREADING####################################################") 
print("\n")
print("Raw T1 Spreading New" + str(spreadT1_New))
print("Raw T1 Spreading Old" + str(spreadT1_Old))
print("Spreading T1Ratio " + str(spreadT1Ratio))
print("\n")
print("Raw T2 Spreading New" + str(spreadT2_New))
print("Raw T2 Spreading Old" + str(spreadT2_Old))
print("Spreading T2Ratio " + str(spreadT2Ratio))
print("\n")
print("Raw T3 Spreading New" + str(spreadT3_New))
print("Raw T3 Spreading Old" + str(spreadT3_Old))
print("Spreading T3Ratio " + str(spreadT3Ratio))
print("\n")

ax2 = fig.add_subplot(222,projection='3d')

if(spreadT1Ratio):
    ax2.bar3d(t1x, t1y, zbot, widths, depths, spreadT1Ratio, shade=True, color='r', label='type1', alpha='1')
if(spreadT2Ratio):
    ax2.bar3d(t2x, t2y, zbot, widths, depths, spreadT2Ratio, shade=True, color='b', label='type2', alpha='1')
if(spreadT3Ratio):
    ax2.bar3d(t3x, t3y, zbot, widths, depths, spreadT3Ratio, shade=True, color='g', label='type3', alpha='1')

ax2.legend([t1_proxy,t2_proxy,t3_proxy], ['type1','type2','type3'])

plt.xlabel('n_trials')
plt.ylabel('Dimensions')
plt.yticks([y+barWidth+1 for y in range(len(t1y))], ['1', '2', '3'])
plt.title('oldSpreadTime/NewSpreadTime')



##################Total FFTW Plan BAR GRAPH####################################################
print("##################Total FFTW Plan####################################################")
print("\n")
print("FFTW Plan T1 New " + str(fftwPlanT1_New))
print("FFTW Plan T1 Old " + str(fftwPlanT1_Old))
print("FFTW Plan T1Ratio " + str(fftwPlanT1Ratio))
print("\n")
print("FFTW Plan T2 New " + str(fftwPlanT2_New))
print("FFTW Plan T2 Old " + str(fftwPlanT2_Old))
print("FFTW Plan T2Ratio " + str(fftwPlanT2Ratio))
print("\n")
print("FFTW Plan T3 New " + str(fftwPlanT3_New))
print("FFTW Plan T3 Old " + str(fftwPlanT3_Old))
print("FFTW Plan T3Ratio " + str(fftwPlanT3Ratio))
print("\n")

ax3 = fig.add_subplot(223,projection='3d')

if(fftwPlanT1Ratio):
    ax3.bar3d(t1x, t1y, zbot, widths, depths, fftwPlanT1Ratio, shade=True, color='r', label='type1', alpha='1')
if(fftwPlanT2Ratio):
    ax3.bar3d(t2x, t2y, zbot, widths, depths, fftwPlanT2Ratio, shade=True, color='b', label='type2', alpha='1')
if(fftwPlanT3Ratio):
    ax3.bar3d(t3x, t3y, zbot, widths, depths, fftwPlanT3Ratio, shade=True, color='g', label='type3', alpha='1')

ax3.legend([t1_proxy,t2_proxy,t3_proxy], ['type1','type2','type3'])

plt.xlabel('n_trials')
plt.ylabel('Dimensions')
plt.yticks([y+barWidth+1 for y in range(len(t1y))], ['1', '2', '3'])
plt.title('totalOldFFtwPlan/NewFftwPlan')


##################Total FFT Exec BAR GRAPH####################################################
print("##################Total FFT Exec####################################################")
print("\n")
print("FFT Exec T1 New " + str(fftT1_New))
print("FFT Exec T1 Old " + str(fftT1_Old))
print("FFT Exec T1Ratio " + str(fftT1Ratio))
print("\n")
print("FFT Exec T2 New " + str(fftT2_New))
print("FFT Exec T2 Old " + str(fftT2_Old))
print("FFT Exec T2Ratio " + str(fftT2Ratio))
print("\n")
print("FFT Exec T3 New " + str(fftT3_New))
print("FFT Exec T3 Old " + str(fftT3_Old))
print("FFT Exec T3Ratio " + str(fftT3Ratio))
print("\n")

ax4 = fig.add_subplot(224,projection='3d')

if(fftT1Ratio):
    ax4.bar3d(t1x, t1y, zbot, widths, depths, fftT1Ratio, shade=True, color='r', label='type1', alpha='1')
if(fftT2Ratio):
    ax4.bar3d(t2x, t2y, zbot, widths, depths, fftT2Ratio, shade=True, color='b', label='type2', alpha='1')
if(fftT3Ratio):
    ax4.bar3d(t3x, t3y, zbot, widths, depths, fftT3Ratio, shade=True, color='g', label='type3', alpha='1')

ax4.legend([t1_proxy,t2_proxy,t3_proxy], ['type1','type2','type3'])

plt.xlabel('n_trials')
plt.ylabel('Dimensions')
plt.yticks([y+barWidth+1 for y in range(len(t1y))], ['1', '2', '3'])
plt.title('totalOldFFtwExec/NewFftwExec')

plt.show()

fig.savefig('timing_breakdowns_rusty_node.png')















import numpy as np
import pytest

from cufinufft import Plan, _compat

import utils

def test_set_nu_raises_on_dtype(to_gpu):
    dtype = np.complex64

    M = 4096
    tol = 1e-3
    shape = (16, 16, 16)
    dim = len(shape)

    kxyz = utils.gen_nu_pts(M, dim=dim).astype(dtype)

    kxyz_gpu = to_gpu(kxyz)

    # Here we'll intentionally contruct an incorrect array dtype.
    kxyz_gpu_wrong_type = to_gpu(kxyz.real.astype(np.float64))

    plan = Plan(1, shape, eps=tol, dtype=dtype)

    with pytest.raises(TypeError):
        plan.setpts(kxyz_gpu_wrong_type[0],
                     kxyz_gpu[1], kxyz_gpu[2])
    with pytest.raises(TypeError):
        plan.setpts(kxyz_gpu[0],
                     kxyz_gpu_wrong_type[1], kxyz_gpu[2])
    with pytest.raises(TypeError):
        plan.setpts(kxyz_gpu[0],
                     kxyz_gpu[1], kxyz_gpu_wrong_type[2])
    with pytest.raises(TypeError):
        plan.setpts(kxyz_gpu_wrong_type[0],
                     kxyz_gpu_wrong_type[1], kxyz_gpu_wrong_type[2])


def test_set_pts_raises_on_size(to_gpu):
    dtype = np.float32
    complex_dtype = np.complex64

    M = 8
    tol = 1e-3
    shape = (16, 16, 16)
    dim = len(shape)

    kxyz = utils.gen_nu_pts(M, dim=dim).astype(dtype)

    kxyz_gpu = to_gpu(kxyz)

    plan = Plan(1, shape, eps=tol, dtype=complex_dtype)

    with pytest.raises(TypeError, match="`y` must be of shape") as err:
        plan.setpts(kxyz_gpu[0], kxyz_gpu[1][:4])

    with pytest.raises(TypeError, match="`z` must be of shape") as err:
        plan.setpts(kxyz_gpu[0], kxyz_gpu[1], kxyz_gpu[2][:4])


def test_set_pts_raises_on_nonvector(to_gpu):
    dtype = np.float32
    complex_dtype = np.complex64

    M = 8
    tol = 1e-3
    shape = (16, 16, 16)
    dim = len(shape)

    kxyz = utils.gen_nu_pts(M, dim=dim).astype(dtype)

    kxyz_gpu = to_gpu(kxyz)

    plan = Plan(1, shape, eps=tol, dtype=complex_dtype)

    with pytest.raises(TypeError, match="`x` must be a vector") as err:
        plan.setpts(kxyz)


def test_set_pts_raises_on_number_of_args(to_gpu):
    dtype = np.float32
    complex_dtype = np.complex64

    M = 8
    tol = 1e-3
    shape = (16,)
    dim = len(shape)

    kxyz = utils.gen_nu_pts(M, dim=3).astype(dtype)

    kxyz_gpu = to_gpu(kxyz)

    plan = Plan(1, shape, eps=tol, dtype=complex_dtype)

    with pytest.raises(TypeError, match="is 1, but `y` was specified") as err:
        plan.setpts(*kxyz_gpu[:2])

    shape = (16, 16)

    plan = Plan(1, shape, eps=tol, dtype=complex_dtype)

    with pytest.raises(TypeError, match="is 2, but `z` was specified") as err:
        plan.setpts(*kxyz_gpu)


def test_wrong_field_names():
    with pytest.raises(TypeError, match="Invalid option 'foo'") as err:
        plan = Plan(1, (8, 8), foo="bar")


def test_exec_raises_on_dtype(to_gpu):
    dtype = np.float32
    complex_dtype = np.complex64

    M = 4096
    tol = 1e-3
    shape = (16, 16, 16)
    dim = len(shape)

    kxyz = utils.gen_nu_pts(M, dim=dim).astype(dtype)
    c = utils.gen_nonuniform_data(M).astype(complex_dtype)
    c_gpu = to_gpu(c)
    # Using c.real gives us wrong dtype here...
    # Need contiguous here since numba does not allow transfers of
    # non-contiguous arrays.
    c_gpu_wrong_dtype = to_gpu(np.ascontiguousarray(c.real))

    kxyz_gpu = to_gpu(kxyz)
    fk_gpu = _compat.array_empty_like(kxyz_gpu, shape, dtype=complex_dtype)
    # Here we'll intentionally contruct an incorrect array dtype.
    fk_gpu_wrong_dtype = _compat.array_empty_like(fk_gpu, shape,
            dtype=np.complex128)

    plan = Plan(1, shape, eps=tol, dtype=complex_dtype)

    plan.setpts(kxyz_gpu[0],
                 kxyz_gpu[1], kxyz_gpu[2])

    with pytest.raises(TypeError):
        plan.execute(c_gpu, fk_gpu_wrong_dtype)

    with pytest.raises(TypeError):
        plan.execute(c_gpu_wrong_dtype, fk_gpu)


def test_dtype_errors():
    with pytest.raises(TypeError, match="Expected complex64 or complex128") as err:
        Plan(1, (8, 8), dtype="uint8")


def test_dtype_warnings():
    with pytest.warns(DeprecationWarning, match="Converting to complex64") as record:
        Plan(1, (8, 8), dtype="float32")

    with pytest.warns(DeprecationWarning, match="Converting to complex128") as record:
        Plan(1, (8, 8), dtype="float64")



from dataclasses import dataclass
import matplotlib.pyplot as plt
import os
import subprocess
import pandas as pd
import numpy as np
import io
from numbers import Number
import time


# function that runs a command line command and returns the output
# it also takes a list of arguments to pass to the command
def run_command(command, args):
    # convert command and args to a string
    try:
        cmd = [command] + args
        print('Running command:', ' '.join(cmd))
        result = subprocess.run(cmd, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        return result.stdout, result.stderr
    except subprocess.CalledProcessError as e:
        print('stdout output:\n', e.stdout)
        print('stderr output:\n', e.stderr)
        print('Error executing command:', e)


def build_args(args):
    args_list = []
    for key, value in args.items():
        args_list.append(key + '=' + value)
    return args_list


# clone the repository
run_command('git', ['clone', 'https://github.com/flatironinstitute/finufft.git'])


def get_cpu_temperature():
    try:
        # Run the sensors command
        output, _ = run_command('sensors', [])
        # Parse the output to find the CPU temperature
        for line in output.split('\n'):
            if 'Core 0' in line:
                # Extract the temperature value
                temp_str = line.split()[2]
                return temp_str
    except subprocess.CalledProcessError as e:
        print('Error executing sensors command:', e)
        return None


all_data = pd.DataFrame()

args = {
    '--type': '1',
    '--prec': 'f',
    '--n_runs': '10',
    '--sort': '1',
    '--N1': '320',
    '--N2': '320',
    '--N3': '1',
    '--ntransf': '1',
    '--thread': '1',
    '--M': '1E6',
    '--tol': '1E-5',
    '--upsampfac': '1.25'
}


@dataclass
class Params:
    prec: str
    N1: Number
    N2: Number
    N3: Number
    ntransf: int
    thread: int
    M: Number
    tol: float


thread_num = int(os.cpu_count())

versions = ['v2.2.0', 'v2.3.0-rc1']
fft_lib = ['fftw', 'ducc']
upsamp = ['1.25', '2.00']
transform = ['1', '2', '3']

ParamList = [
    Params('f', 1e4, 1, 1, 1, 1, 1e7, 1e-4),
    Params('d', 1e4, 1, 1, 1, 1, 1e7, 1e-9),
    Params('f', 320, 320, 1, 1, 1, 1e7, 1e-5),
    Params('d', 320, 320, 1, 1, 1, 1e7, 1e-9),
    Params('f', 320, 320, 1, thread_num, thread_num, 1e7, 1e-5),
    Params('d', 192, 192, 128, 1, thread_num, 1e7, 1e-7),
]


def plot_stacked_bar_chart(pivot_data, speedup_data, args, figname):
    categories = list(pivot_data.keys())
    versions = list(pivot_data[categories[0]].keys())

    # Create a figure
    plt.figure(figsize=(7.5, 7.5))

    # Initialize the bottom array for stacking
    bottom = np.zeros(len(versions))

    # Plot each category
    for category in categories:
        values = [pivot_data[category][version] for version in versions]
        plt.bar(versions, values, bottom=bottom, label=category)
        bottom += np.array(values)

    # Add speedup annotations
    for i, version in enumerate(versions):
        speedup = \
            speedup_data[(speedup_data['version'] == version) & (speedup_data['event'] == 'amortized')][
                'speedup'].values[0]
        plt.text(i, bottom[i], f'{speedup:.2f}x', ha='center', va='bottom')

    # Add labels and title
    plt.ylabel('time (ms)')
    plt.xlabel('version')
    plt.xticks(rotation=0)

    # for txt create a string with all the arguments insert a newline after half of the arguments
    txt = ''
    for i, (key, value) in enumerate(args.items()):
        txt += f'{key[2:]}:{value}'
        if i == len(args) // 2:
            txt += '\n'
        else:
            txt += ' '

    plt.title(txt)
    plt.legend(loc='lower left')

    # Save the figure
    plt.savefig(figname + '.png')
    plt.show()


for version in versions:
    for fft in fft_lib:
        if version == 'v2.2.0' and fft == 'ducc':
            continue
        run_command('git', ['-C', 'finufft', 'reset', '--hard'])
        run_command('git', ['-C', 'finufft', 'checkout', version])
        # checkout folder perftest from master branch
        run_command('git', ['-C', 'finufft', 'checkout', 'origin/master', '--', 'perftest'])
        # run_command('rm', ['-rf', 'build'])
        if fft == 'ducc':
            run_command('cmake',
                        ['-S', 'finufft', '-B', 'build', '-DCMAKE_BUILD_TYPE=Release', '-DFINUFFT_BUILD_TESTS=ON',
                         '-DFINUFFT_USE_DUCC0=ON'])
        else:
            run_command('cmake',
                        ['-S', 'finufft', '-B', 'build', '-DCMAKE_BUILD_TYPE=Release', '-DFINUFFT_BUILD_TESTS=ON',
                         '-DFINUFFT_FFTW_LIBRARIES=DOWNLOAD'])
        run_command('cmake', ['--build', 'build', '-j', str(os.cpu_count()), '--target', 'perftest'])
        for param in ParamList:
            for key, value in param.__dict__.items():
                args['--' + key] = str(value)
            for type in transform:
                args['--' + 'type'] = type
                for upsampfac in upsamp:
                    args['--upsampfac'] = upsampfac
                    # while (cpu_temp := float(get_cpu_temperature()[:-2])) > 44.0:
                    #     print(f'CPU temperature is {cpu_temp}°C, waiting for it to cool down...')
                    #     time.sleep(5)
                    if param.thread == 1:
                        out, _ = run_command('taskset', ['-c', '0', 'build/perftest/perftest'] + build_args(args))
                    else:
                        time.sleep(30)
                        out, _ = run_command('build/perftest/perftest', build_args(args))
                    # parse the output, escape all the lines that start with #
                    out = io.StringIO(out)
                    lines = out.readlines()
                    conf = [line for line in lines if line.startswith('#')]
                    print(*conf, sep='')
                    stdout = '\n'.join([line for line in lines if not line.startswith('#')])
                    # convert stdout to a dataframe from csv string
                    dt = pd.read_csv(io.StringIO(stdout), sep=',')
                    # add columns with version and configuration
                    dt['version'] = version[1:] + '-' + fft
                    for key, value in args.items():
                        dt[key[2:]] = value
                    all_data = pd.concat((all_data, dt), ignore_index=True)
                    print(dt)

# Replace the amortized event in all_data
all_data.loc[all_data['event'] == 'amortized', 'min(ms)'] = (
        all_data[all_data['event'] == 'makeplan']['min(ms)'].values
        + all_data[all_data['event'] == 'setpts']['min(ms)'].values
        + all_data[all_data['event'] == 'execute']['min(ms)'].values
)

print(all_data)
all_data.to_csv('all_data.csv')
for param in ParamList:
    for key, value in param.__dict__.items():
        args['--' + key] = str(value)
    for upsampfac in upsamp:
        args['--upsampfac'] = upsampfac
        for type in transform:
            args['--' + 'type'] = type
            this_data = all_data
            # select data for this specific upsampling factor, type and parameters
            for key, value in args.items():
                this_data = this_data[this_data[key[2:]] == value]
            print(this_data)
            name = f'{int(param.N1)}x{int(param.N2)}x{int(param.N3)}-type-{type}-upsamp{upsampfac}-prec{param.prec}-thread{int(param.thread)}'
            # select the baseline data
            baseline = this_data[this_data['version'] == '2.2.0-fftw']
            # calculate the amortized time for the baseline
            baseline_amortized = baseline[baseline['event'] == 'amortized']['min(ms)'].values[0]
            # calculate the speedup for all the other versions
            this_data['speedup'] = baseline_amortized / this_data[this_data['event'] == 'amortized']['min(ms)']
            # pivot the data
            pivot = this_data.pivot(index='version', columns='event', values='min(ms)')
            pivot = pivot.drop(columns='amortized')
            # plot the stacked bar chart
            plot_stacked_bar_chart(pivot, this_data, args, name)

# this_data = all_data[all_data['upsampfac'] == upsampfac]
# baseline = this_data[this_data['version'] == '2.2.0-fftw']
# baseline_amortized = baseline[baseline['event'] == 'amortized']['mean(ms)'].values[0]
# this_data['speedup'] = baseline_amortized / this_data[this_data['event'] == 'amortized']['mean(ms)']
# pivot = this_data.pivot(index='version', columns='event', values='mean(ms)')
# pivot = pivot.drop(columns='amortized')
# plot_stacked_bar_chart(pivot, this_data, args, '2d-' + str(upsampfac))

if __name__ == '__main__':
    pass
